""" Module to parse output from BBDuk """

import logging
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph
from multiqc.utils import config

log = logging.getLogger(__name__)

VERSION_REGEX = r"Version ([\d\.]+)"


class MultiqcModule(BaseMultiqcModule):
    """BBDuk Module"""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="BBDuk",
            anchor="bbduk",
            href="https://jgi.doe.gov/data-and-tools/software-tools/bbtools/bb-tools-user-guide/bbduk-guide/",
            info="""is a tool performing common data-quality-related trimming,
            filtering, and masking operations with a kmer based approach""",
            # One publication, but only for the merge tool:
            # doi="10.1371/journal.pone.0185056",
        )

        # Define the main bbduk multiqc data object
        self.bbduk_data = dict()

        for f in self.find_log_files("bbduk", filehandles=True):
            self.parse_logs(f)

        self.bbduk_data = self.ignore_samples(self.bbduk_data)

        if len(self.bbduk_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.bbduk_data)} reports")

        # Write data to file
        self.write_data_file(self.bbduk_data, "bbduk")

        self.bbduk_general_stats()
        self.bbduk_bargraph_plot()

    def parse_logs(self, f):
        """Parses a BBDuk stdout saved in a file"""

        s_name = f["s_name"]
        for line in f["f"]:
            if "jgi.BBDuk" in line and "in1=" in line:
                s_name = line.split("in1=")[1].split(" ")[0]
                s_name = self.clean_s_name(s_name, f)

            if line.startswith("Version"):
                version_match = re.search(VERSION_REGEX, line)
                if version_match:
                    self.add_software_version(version_match.group(1), s_name)

            if "Input:" in line:
                matches = re.search(r"Input:\s+(\d+) reads\s+(\d+) bases", line)
                if matches:
                    self.add_data_source(f, s_name)
                    if s_name in self.bbduk_data:
                        log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                    self.bbduk_data[s_name] = dict()

                    self.bbduk_data[s_name]["Input reads"] = int(matches.group(1))
                    self.bbduk_data[s_name]["Input bases"] = int(matches.group(2))
            # Don't start using regexes until we're in that block
            elif "Input reads" in self.bbduk_data.get(s_name, {}):
                cats = [
                    "QTrimmed",
                    "KTrimmed",
                    "Trimmed by overlap",
                    "Low quality discards",
                    "Low entropy discards",
                    "Total Removed",
                    "Result",
                ]
                for cat in cats:
                    matches = re.search(rf"{cat}:\s+(\d+) reads \(([\d\.]+)%\)\s+(\d+) bases \(([\d\.]+)%\)", line)
                    if matches:
                        self.bbduk_data[s_name][cat + " reads"] = int(matches.group(1))
                        self.bbduk_data[s_name][cat + " reads percent"] = float(matches.group(2))
                        self.bbduk_data[s_name][cat + " bases"] = int(matches.group(3))
                        self.bbduk_data[s_name][cat + " bases percent"] = float(matches.group(4))
                        break
            elif "Reads Processed:" in line:
                return

    def bbduk_general_stats(self):
        """BBDuk read counts for general stats"""
        headers = {
            "Total Removed bases percent": {
                "title": "Bases Removed (%)",
                "description": "Percentage of bases removed after filtering",
                "scale": "YlOrBr",
                "max": 100,
            },
            "Total Removed bases": {
                "title": f"Bases Removed ({config.base_count_prefix})",
                "description": f"Total Bases removed ({config.base_count_desc})",
                "scale": "Reds",
                "shared_key": "base_count",
                "modify": lambda x: x * config.base_count_multiplier,
                "hidden": True,
            },
            "Total Removed reads percent": {
                "title": "Reads Removed (%)",
                "description": "Percentage of reads removed after filtering",
                "scale": "OrRd",
                "max": 100,
            },
            "Total Removed reads": {
                "title": f"Reads Removed ({config.read_count_prefix})",
                "description": f"Total Reads removed ({config.read_count_desc})",
                "scale": "Reds",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
                "hidden": True,
            },
            "Input reads": {
                "title": f"Total Input Reads ({config.read_count_prefix})",
                "description": f"Total number of input reads to BBDuk ({config.read_count_desc})",
                "scale": "Greens",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
                "hidden": True,
            },
        }

        self.general_stats_addcols(self.bbduk_data, headers)

    def bbduk_bargraph_plot(self):
        """
        Beeswarm displaying all possible filtering results reported by BBDuk.

        We don't display this as a barchart as the total across all categories
        of filters reported don't match exactly the total reads remaining (I
        assume there is additional default filtering carried out)
        """
        cats = [
            "Result",
            "QTrimmed",
            "KTrimmed",
            "Trimmed by overlap",
            "Low quality discards",
            "Low entropy discards",
        ]
        pconfig = {
            "id": "bbduk-filtered-barplot",
            "title": "BBDuk: Filtered reads",
            "ylab": "Number of Reads",
            "data_labels": [
                {"name": "Reads", "ylab": "Number of Reads"},
                {"name": "Bases", "ylab": "Number of Base Pairs"},
            ],
        }

        self.add_section(
            name="BBDuk: Filtered Reads",
            anchor="bbduk-filtered-reads",
            description="The number of reads removed by various BBDuk filters",
            plot=bargraph.plot(
                [self.bbduk_data, self.bbduk_data],
                [
                    [f"{cat} reads" for cat in cats],
                    [f"{cat} bases" for cat in cats],
                ],
                pconfig,
            ),
        )



""" MultiQC module to parse output from NanoStat """


import logging

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, table
from multiqc.utils import mqc_colour

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """NanoStat module"""

    _KEYS_MAPPING = {
        "number_of_reads": "Number of reads",
        "number_of_bases": "Total bases",
        "number_of_bases_aligned": "Total bases aligned",
        "fraction_bases_aligned": "Fraction of bases aligned",
        "median_read_length": "Median read length",
        "mean_read_length": "Mean read length",
        "read_length_stdev": "STDEV read length",
        "n50": "Read length N50",
        "average_identity": "Average percent identity",
        "median_identity": "Median percent identity",
        "active_channels": "Active channels",
        "mean_qual": "Mean read quality",
        "median_qual": "Median read quality",
    }

    _KEYS_READ_Q = [
        ">Q5",
        ">Q7",
        ">Q10",
        ">Q12",
        ">Q15",
    ]
    _stat_types = ("aligned", "seq summary", "fastq", "fasta", "unrecognized")

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="NanoStat",
            anchor="nanostat",
            href="https://github.com/wdecoster/nanostat/",
            info="reports various statistics from a long read sequencing dataset in FASTQ, BAM or sequencing summary format.",
            doi="10.1093/bioinformatics/bty149",
        )

        # Find and load any NanoStat reports
        self.nanostat_data = dict()
        self.has_qscores = False
        self.has_aligned = False
        self.has_seq_summary = False
        self.has_fastq = False
        self.has_fasta = False
        for f in self.find_log_files("nanostat", filehandles=True):
            self.parse_nanostat_log(f)

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, f["s_name"])

        for f in self.find_log_files("nanostat/legacy", filehandles=True):
            self.parse_legacy_nanostat_log(f)

        # Filter to strip out ignored sample names
        self.nanostat_data = self.ignore_samples(self.nanostat_data)

        if len(self.nanostat_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.nanostat_data)} reports")

        # Write parsed report data to a file
        self.write_data_file(self.nanostat_data, "multiqc_nanostat")

        # Stats Tables
        if self.has_aligned:
            self.nanostat_stats_table("aligned")
        if self.has_seq_summary:
            self.nanostat_stats_table("seq summary")
        if self.has_fastq:
            self.nanostat_stats_table("fastq")
        if self.has_fasta:
            self.nanostat_stats_table("fasta")

        # Quality distribution Plot
        if self.has_qscores:
            self.reads_by_quality_plot()

    def parse_nanostat_log(self, f):
        """Parse output from NanoStat

        Note: Tool can be run in two different modes, giving two variants to the output.
        To avoid overwriting keys from different modes, keys are given a suffix.
        """
        nano_stats = {}
        for line in f["f"]:
            parts = line.strip().split()
            if len(parts) == 2 and parts[0] in self._KEYS_MAPPING.keys():
                key = self._KEYS_MAPPING.get(parts[0])
                if key:
                    nano_stats[key] = float(parts[1])
            else:
                parts = line.strip().split(":")
                key = parts[0].replace("Reads ", "")
                if key in self._KEYS_READ_Q:
                    # Number of reads above Q score cutoff
                    val = int(parts[1].strip().split()[0])
                    nano_stats[key] = val
        self.save_data(f, nano_stats)

    def parse_legacy_nanostat_log(self, f):
        """Parse legacy output from NanoStat

        Note: Tool can be run in two different modes, giving two variants to the output.
        To avoid overwriting keys from different modes, keys are given a suffix.
        """
        nano_stats = {}
        for line in f["f"]:
            parts = line.strip().split(":")
            if len(parts) == 0:
                continue

            key = parts[0]

            if key in self._KEYS_MAPPING.values():
                val = float(parts[1].replace(",", ""))
                nano_stats[key] = val
            elif key in self._KEYS_READ_Q:
                # Number of reads above Q score cutoff
                val = int(parts[1].strip().split()[0])
                nano_stats[key] = val
        self.save_data(f, nano_stats)

    def save_data(self, f, nano_stats):
        """
        Normalise fields and save parsed data.

        Used for both legacy and new data formats.
        """
        if ">Q5" in nano_stats:
            self.has_qscores = True

        if "Total bases aligned" in nano_stats:
            stat_type = "aligned"
            self.has_aligned = True
        elif "Active channels" in nano_stats:
            stat_type = "seq summary"
            self.has_seq_summary = True
        elif "Mean read quality" in nano_stats:
            stat_type = "fastq"
            self.has_fastq = True
        elif "Mean read length" in nano_stats:
            stat_type = "fasta"
            self.has_fasta = True
        else:
            log.debug(f"Did not recognise NanoStat file '{f['fn']}' - skipping")
            return

        out_d = {f"{k}_{stat_type}": v for k, v in nano_stats.items()}

        # Warn if we find overlapping data for the same sample
        if f["s_name"] in self.nanostat_data:
            # Only if the same has some keys in common
            if not set(self.nanostat_data[f["s_name"]].keys()).isdisjoint(out_d.keys()):
                log.debug(f"Duplicate sample data found! Overwriting: {f['s_name']}")

        self.nanostat_data.setdefault(f["s_name"], {}).update(out_d)

        self.add_data_source(f)

    def nanostat_stats_table(self, stat_type):
        """Take the parsed stats from the Kallisto report and add it to the
        basic stats table at the top of the report"""

        headers_base = {
            "Active channels": {
                "title": "Active channels",
                "description": "Active channels",
                "scale": "Greens",
                "format": "{:,.0f}",
            },
            "Median read length": {
                "title": "Median length",
                "description": "Median read length (bp)",
                "suffix": " bp",
                "format": "{:,.0f}",
                "shared_key": "nucleotides",
                "scale": "BuPu",
            },
            "Mean read length": {
                "title": "Mean length",
                "description": "Mean read length (bp)",
                "suffix": " bp",
                "scale": "Purples",
                "format": "{:,.0f}",
                "shared_key": "nucleotides",
                "hidden": True,
            },
            "Read length N50": {
                "title": "Read N50",
                "description": "Read length N50",
                "format": "{:,.0f}",
                "suffix": " bp",
                "scale": "RdPu",
            },
            "Median read quality": {
                "title": "Median Qual",
                "description": "Median read quality (Phred scale)",
                "shared_key": "phred_score",
                "scale": "RdYlGn",
            },
            "Mean read quality": {
                "title": "Mean Qual",
                "description": "Mean read quality (Phred scale)",
                "scale": "PiYG",
                "shared_key": "phred_score",
                "hidden": True,
            },
            "Median percent identity": {
                "title": "Median Identity",
                "description": "Median percent identity",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "RdYlBu",
                "shared_key": "percent_identity",
            },
            "Average percent identity": {
                "title": "Mean Identity",
                "description": "Average percent identity",
                "max": 100,
                "suffix": "%",
                "scale": "Spectral",
                "shared_key": "percent_identity",
                "hidden": True,
            },
            "Number of reads": {
                "title": f"# Reads ({config.long_read_count_prefix})",
                "description": f"Number of reads ({config.long_read_count_desc})",
                "modify": lambda x: x * config.long_read_count_multiplier,
                "shared_key": "long_read_count",
                "scale": "YlGn",
            },
            "Total bases": {
                "title": f"Total Bases ({config.base_count_prefix})",
                "description": f"Total bases ({config.base_count_desc})",
                "modify": lambda x: x * config.base_count_multiplier,
                "shared_key": "base_count",
                "scale": "BrBG",
            },
            "Total bases aligned": {
                "title": f"Aligned Bases ({config.base_count_prefix})",
                "description": f"Total bases aligned ({config.base_count_desc})",
                "modify": lambda x: x * config.base_count_multiplier,
                "shared_key": "base_count",
                "scale": "PuOr",
            },
        }

        # Add the stat_type suffix
        headers = {}
        for k in headers_base:
            key = f"{k}_{stat_type}"
            headers[key] = headers_base.get(k, dict()).copy()

        # Table config
        table_config = {
            "namespace": "NanoStat",
            "id": f"nanostat_{stat_type.replace(' ', '_')}_stats_table",
            "table_title": f"NanoStat {stat_type}",
        }

        # Add the report section
        self.add_section(
            name="Summary Statistics",
            anchor=f"nanostat_{stat_type.replace(' ', '_')}_stats",
            plot=table.plot(self.nanostat_data, headers, table_config),
        )

    def reads_by_quality_plot(self):
        """Make the HighCharts HTML to plot the reads by quality"""

        def _get_total_reads(data_dict):
            stat_type = self._stat_types[0]
            for stat_type in self._stat_types:
                total_key = f"Number of reads_{stat_type}"
                if total_key in data_dict:
                    return data_dict[total_key], stat_type
            return None, None

        bar_data = {}
        stat_type = "unrecognized"
        # Order of keys, from >Q5 to >Q15
        _range_names = {
            ">Q5": "<Q5",
            ">Q7": "Q5-7",
            ">Q10": "Q7-10",
            ">Q12": "Q10-12",
            ">Q15": "Q12-15",
            "rest": ">Q15",
        }
        for s_name, data_dict in self.nanostat_data.items():
            reads_total, stat_type = _get_total_reads(data_dict)
            if stat_type == "fasta":
                log.debug(f"Sample '{s_name}' has no quality metrics - excluded from quality plot")
                continue
            if s_name in bar_data and stat_type == "aligned":
                log.debug(f"Sample '{s_name}' duplicated in the quality plot - ignoring aligned data")
                continue
            elif s_name in bar_data and stat_type == "seq summary":
                log.debug(f"Sample '{s_name}' duplicated in the quality plot - overwriting with seq summary data")
            bar_data[s_name] = {}

            prev_reads = reads_total
            for k, range_name in _range_names.items():
                if k != "rest":
                    data_key = f"{k}_{stat_type}"
                    reads_gt = data_dict[data_key]

                    bar_data[s_name][range_name] = prev_reads - reads_gt

                    if bar_data[s_name][range_name] < 0:
                        log.error(f"Error on {s_name} {range_name} {data_key} . Negative number of reads")
                    prev_reads = reads_gt
                else:
                    data_key = f">Q15_{stat_type}"
                    bar_data[s_name][range_name] = data_dict[data_key]

        cats = {}
        keys = reversed(list(_range_names.values()))
        colours = mqc_colour.mqc_colour_scale("RdYlGn-rev", 0, len(_range_names))
        for idx, k in enumerate(keys):
            cats[k] = {"name": "Reads " + k, "color": colours.get_colour(idx, lighten=1)}

        # Config for the plot
        config = {
            "id": "nanostat_quality_dist",
            "title": "NanoStat: Reads by quality",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
        }

        # Add the report section
        self.add_section(
            name="Reads by quality",
            anchor="nanostat_read_qualities",
            description="Read counts categorised by read quality (Phred score).",
            helptext="""
            Sequencing machines assign each generated read a quality score using the
            [Phred scale](https://en.wikipedia.org/wiki/Phred_quality_score).
            The phred score represents the liklelyhood that a given read contains errors.
            High quality reads have a high score.
            """,
            plot=bargraph.plot(bar_data, cats, config),
        )



""" MultiQC submodule to parse output from Picard BaseDistributionByCycleMetrics"""

import logging
from collections import defaultdict

from multiqc.modules.picard import util
from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(module):
    """Find Picard BaseDistributionByCycleMetrics reports and parse their data"""

    data_by_sample = dict()
    samplestats_by_sample = dict()

    # Go through logs and find Metrics
    for f in module.find_log_files("picard/basedistributionbycycle", filehandles=True):
        # Sample name from input file name by default.
        s_name = f["s_name"]

        # A file can be concatenated from multiple samples, so we need to keep track of
        # the current sample name and header.
        keys = None
        data_by_read_end = defaultdict(dict)
        max_cycle_r1 = 0

        def _finalize_sample(data_by_read_end, s_name: str):
            """
            Populate `data_by_sample` and `samplestats_by_sample`
            """
            # set up the set of s_names
            if 2 in set(data_by_read_end):
                s_names = {1: f"{s_name}_R1", 2: f"{s_name}_R2"}
            else:
                s_names = {1: s_name}

            for read_end in s_names:
                data_by_cycle = data_by_read_end[read_end]
                s_name = s_names[read_end]
                data_by_sample[s_name] = data_by_cycle
                if s_name in data_by_sample:
                    log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")
                module.add_data_source(f, s_name, section="BaseDistributionByCycle")
                sample_stats = {
                    "sum_pct_a": 0,
                    "sum_pct_c": 0,
                    "sum_pct_g": 0,
                    "sum_pct_t": 0,
                    "sum_pct_n": 0,
                    "cycle_count": 0,
                }
                samplestats_by_sample[s_name] = sample_stats
                for c, row in data_by_cycle.items():
                    pct_a, pct_c, pct_g, pct_t, pct_n = row
                    sample_stats["sum_pct_a"] += pct_a
                    sample_stats["sum_pct_c"] += pct_c
                    sample_stats["sum_pct_g"] += pct_g
                    sample_stats["sum_pct_t"] += pct_t
                    sample_stats["sum_pct_n"] += pct_n
                sample_stats["cycle_count"] += len(data_by_cycle.keys())

        for line in f["f"]:
            maybe_s_name = util.extract_sample_name(module, line, f, picard_tool="CollectBaseDistributionByCycle")
            if maybe_s_name:
                # Starts information for a new sample
                s_name = maybe_s_name
                keys = None

            if s_name is None:
                continue

            if util.is_line_right_before_table(line, picard_class="BaseDistributionByCycleMetrics"):
                keys = f["f"].readline().strip("\n").split("\t")
                assert keys == ["READ_END", "CYCLE", "PCT_A", "PCT_C", "PCT_G", "PCT_T", "PCT_N"]

            elif keys:
                raw_vals = line.strip("\n").split("\t")
                if len(raw_vals) != len(keys):
                    # Finalize previous sample
                    if data_by_read_end:
                        _finalize_sample(data_by_read_end, s_name)
                    # Reset for next sample
                    max_cycle_r1 = 0
                    s_name = None
                    data_by_read_end = defaultdict(dict)
                    continue

                vals = []
                for v in raw_vals:
                    try:
                        v = float(v)
                    except ValueError:
                        pass
                    vals.append(v)

                read_end, cycle, pct_a, pct_c, pct_g, pct_t, pct_n = vals
                try:
                    cycle = int(cycle)
                except ValueError:
                    continue
                if read_end == 1.0:
                    max_cycle_r1 = max(max_cycle_r1, cycle)
                else:
                    cycle -= max_cycle_r1
                data_by_read_end[read_end][cycle] = (pct_a, pct_c, pct_g, pct_t, pct_n)

        if data_by_read_end and s_name:
            _finalize_sample(data_by_read_end, s_name)

    # Filter to strip out ignored sample names
    data_by_sample = module.ignore_samples(data_by_sample)
    samplestats_by_sample = module.ignore_samples(samplestats_by_sample)
    if len(data_by_sample) == 0:
        return 0

    # Calculate summed mean values for all read orientations
    for s_name, v in samplestats_by_sample.items():
        v["mean_pct_a"] = v["sum_pct_a"] / v["cycle_count"]
        v["mean_pct_c"] = v["sum_pct_c"] / v["cycle_count"]
        v["mean_pct_g"] = v["sum_pct_g"] / v["cycle_count"]
        v["mean_pct_t"] = v["sum_pct_t"] / v["cycle_count"]

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    module.add_software_version(None)

    # Write parsed data to a file
    module.write_data_file(samplestats_by_sample, "multiqc_picard_baseContent")

    # Plot the data and add section
    pconfig = {
        "id": "picard_base_distribution_by_cycle",
        "title": "Picard: Base Distribution",
        "ylab": "%",
        "xlab": "Cycle #",
        "xDecimals": False,
        "tt_label": "<b>cycle {point.x}</b>: {point.y:.2f} %",
        "ymax": 100,
        "ymin": 0,
        "data_labels": [
            {"name": "% Adenine", "ylab": "% Adenine"},
            {"name": "% Cytosine", "ylab": "% Cytosine"},
            {"name": "% Guanine", "ylab": "% Guanine"},
            {"name": "% Thymine", "ylab": "% Thymine"},
            {"name": "% Undetermined", "ylab": "% Undetermined"},
        ],
    }

    # build list of linegraphs
    linegraph_data = [{}, {}, {}, {}, {}]
    for s_name, cycles in data_by_sample.items():
        for lg, index in zip(linegraph_data, range(5)):
            lg[s_name] = {cycle: tup[index] for cycle, tup in cycles.items()}

    module.add_section(
        name="Base Distribution",
        anchor="picard-base-distribution-by-cycle",
        description="Plot shows the distribution of bases by cycle.",
        plot=linegraph.plot(linegraph_data, pconfig),
    )

    # Return the number of detected samples to the parent module
    return len(data_by_sample)



""" MultiQC submodule to parse output from Picard InsertSizeMetrics """

import logging

from multiqc import config
from multiqc.modules.picard import util
from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(module):
    """Find Picard InsertSizeMetrics reports and parse their data"""

    data_by_sample = dict()
    histogram_by_sample = dict()
    samplestats_by_sample = dict()

    # Go through logs and find Metrics
    for f in module.find_log_files("picard/insertsize", filehandles=True):
        # Sample name from input file name by default
        s_name = f["s_name"]
        in_hist = False

        for line in f["f"]:
            maybe_s_name = util.extract_sample_name(
                module,
                line,
                f,
                picard_tool="CollectInsertSizeMetrics",
                sentieon_algo="InsertSizeMetricAlgo",
            )
            if maybe_s_name:
                s_name = maybe_s_name

            if s_name is None:
                continue

            # Catch the histogram values
            if in_hist:
                try:
                    sections = line.split("\t")
                    ins = int(sections[0])
                    tot_count = sum([int(x) for x in sections[1:]])
                    histogram_by_sample[s_name][ins] = tot_count
                    samplestats_by_sample[s_name]["total_count"] += tot_count
                except ValueError:
                    # Reset in case we have more in this log file
                    s_name = None
                    in_hist = False

            if util.is_line_right_before_table(
                line, picard_class="InsertSizeMetrics", sentieon_algo="InsertSizeMetricAlgo"
            ):
                keys = f["f"].readline().strip("\n").split("\t")
                vals = f["f"].readline().strip("\n").split("\t")
                if len(vals) != len(keys):
                    continue

                if s_name in data_by_sample:
                    log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")

                module.add_data_source(f, s_name, section="InsertSizeMetrics")
                samplestats_by_sample[s_name] = {"total_count": 0, "meansum": 0, "total_pairs": 0}
                orientation_idx = keys.index("PAIR_ORIENTATION")

                while len(vals) == len(keys):
                    pair_orientation = vals[orientation_idx]
                    rowkey = f"{s_name}_{pair_orientation}"
                    data_by_sample[rowkey] = dict()
                    data_by_sample[rowkey]["SAMPLE_NAME"] = s_name
                    for i, k in enumerate(keys):
                        try:
                            data_by_sample[rowkey][k] = float(vals[i])
                        except ValueError:
                            try:
                                unfixed = vals[i]
                                fixed = unfixed.replace(",", ".")
                                data_by_sample[rowkey][k] = float(fixed)
                                log.debug(f"Switching commas for points in '{f['fn']}': {unfixed} -> {fixed}")
                            except ValueError:
                                data_by_sample[rowkey][k] = vals[i]
                        except IndexError:
                            pass  # missing data
                    # Add to mean sums
                    rp = data_by_sample[rowkey]["READ_PAIRS"]
                    mis = data_by_sample[rowkey]["MEAN_INSERT_SIZE"]
                    samplestats_by_sample[s_name]["meansum"] += rp * mis
                    samplestats_by_sample[s_name]["total_pairs"] += rp

                    vals = f["f"].readline().strip("\n").split("\t")

            if line.startswith("## HISTOGRAM"):
                keys = f["f"].readline().strip("\n").split("\t")
                assert len(keys) >= 2, (keys, f)
                in_hist = True
                histogram_by_sample[s_name] = dict()

    # Calculate summed mean values for all read orientations
    for s_name, v in samplestats_by_sample.items():
        try:
            samplestats_by_sample[s_name]["summed_mean"] = v["meansum"] / v["total_pairs"]
        except ZeroDivisionError:
            # The mean of zero elements is zero
            samplestats_by_sample[s_name]["summed_mean"] = 0

    # Calculate summed median values for all read orientations
    for s_name in histogram_by_sample:
        j = 0
        for idx, c in histogram_by_sample[s_name].items():
            j += c
            if j > (samplestats_by_sample[s_name]["total_count"] / 2):
                samplestats_by_sample[s_name]["summed_median"] = idx
                break

    # Filter to strip out ignored sample names
    data_by_sample = module.ignore_samples(data_by_sample)
    if len(data_by_sample) == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    module.add_software_version(None)

    # Write parsed data to a file
    module.write_data_file(data_by_sample, f"multiqc_{module.anchor}_insertSize")

    # Do we have median insert sizes?
    missing_medians = False
    for v in samplestats_by_sample.values():
        if "summed_median" not in v:
            missing_medians = True

    # Add to general stats table
    headers = {
        "summed_median": {
            "title": "Insert Size",
            "description": "Median Insert Size, all read orientations (bp)",
            "min": 0,
            "suffix": " bp",
            "format": "{:,.0f}",
            "scale": "GnBu",
        },
        "summed_mean": {
            "title": "Mean Insert Size",
            "description": "Mean Insert Size, all read orientations (bp)",
            "min": 0,
            "suffix": " bp",
            "format": "{:,.0f}",
            "scale": "GnBu",
            "hidden": False if missing_medians else True,
        },
    }
    module.general_stats_addcols(samplestats_by_sample, headers, namespace="InsertSizeMetrics")

    # Section with histogram plot
    if len(histogram_by_sample) > 0:
        # Make a normalised percentage version of the data
        data_percent = {}
        for s_name, data in histogram_by_sample.items():
            data_percent[s_name] = dict()
            total = float(sum(data.values()))
            for k, v in data.items():
                data_percent[s_name][k] = (v / total) * 100

        # Allow customisation of how smooth the plot is
        try:
            insertsize_smooth_points = int(config.picard_config["insertsize_smooth_points"])
            log.debug(f"Custom Picard insert size smoothing: {insertsize_smooth_points}")
        except (AttributeError, KeyError, ValueError):
            insertsize_smooth_points = 500

        # Plot the data and add section
        pconfig = {
            "smooth_points": insertsize_smooth_points,
            "smooth_points_sumcounts": [True, False],
            "id": f"{module.anchor}_insert_size",
            "title": f"{module.name}: Insert Size",
            "ylab": "Count",
            "xlab": "Insert Size (bp)",
            "xDecimals": False,
            "tt_label": "<b>{point.x} bp</b>: {point.y:.0f}",
            "ymin": 0,
            "data_labels": [
                {"name": "Counts", "ylab": "Coverage"},
                {"name": "Percentages", "ylab": "Percentage of Counts"},
            ],
        }
        try:
            pconfig["xmax"] = config.picard_config["insertsize_xmax"]
        except (AttributeError, KeyError):
            pass

        module.add_section(
            name="Insert Size",
            anchor=f"{module.anchor}-insertsize",
            description="Plot shows the number of reads at a given insert size. Reads "
            "with different orientations are summed.",
            plot=linegraph.plot([histogram_by_sample, data_percent], pconfig),
        )

    # Return the number of detected samples to the parent module
    return len(data_by_sample)



""" MultiQC module to parse output from featureCounts """


import logging
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="featureCounts",
            anchor="featurecounts",
            target="Subread featureCounts",
            href="http://subread.sourceforge.net/",
            info="is a highly efficient general-purpose read summarization program"
            " that counts mapped reads for genomic features such as genes, exons,"
            " promoter, gene bodies, genomic bins and chromosomal locations.",
            doi="10.1093/bioinformatics/btt656",
        )

        # Find and load any featureCounts reports
        self.featurecounts_data = dict()
        self.featurecounts_keys = list()
        for f in self.find_log_files("featurecounts"):
            self.parse_featurecounts_report(f)

        # Filter to strip out ignored sample names
        self.featurecounts_data = self.ignore_samples(self.featurecounts_data)

        if len(self.featurecounts_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.featurecounts_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file
        self.write_data_file(self.featurecounts_data, "multiqc_featureCounts")

        # Basic Stats Table
        self.featurecounts_stats_table()

        # Assignment bar plot
        self.add_section(plot=self.featureCounts_chart())

    def parse_featurecounts_report(self, f):
        """Parse the featureCounts log file."""

        file_names = list()
        parsed_data = dict()
        split_sep = "\t"
        for line in f["f"].splitlines():
            thisrow = list()

            # If this is from Rsubread then the formatting can be very variable
            # Default search pattern is quite generic, so f
            if len(file_names) == 0 and len(line.split(split_sep)) < 2:
                # Split by whitespace and strip quote marks
                # NB: Will break if sample names have whitespace. RSubread output is so variable that this is difficult to avoid
                split_sep = None

            s = line.split(split_sep)
            s = [sv.strip('"') for sv in s]

            if len(s) < 2:
                continue
            if s[0] == "Status":
                for f_name in s[1:]:
                    file_names.append(f_name)
            elif len(file_names) + 1 == len(s):
                k = s[0]
                if k not in self.featurecounts_keys:
                    self.featurecounts_keys.append(k)
                for val in s[1:]:
                    try:
                        thisrow.append(int(val))
                    except ValueError:
                        pass
            if len(thisrow) > 0:
                parsed_data[k] = thisrow

        # Check that this actually is a featureCounts file, as format and parsing is quite general
        if "Assigned" not in parsed_data.keys():
            return None

        for idx, f_name in enumerate(file_names):
            # Clean up sample name
            s_name = self.clean_s_name(f_name, f)

            # Reorganised parsed data for this sample
            # Collect total count number
            data = dict()
            data["Total"] = 0
            for k in parsed_data:
                data[k] = parsed_data[k][idx]
                data["Total"] += parsed_data[k][idx]

            # Calculate the percent aligned if we can
            try:
                data["percent_assigned"] = (float(data["Assigned"]) / float(data["Total"])) * 100.0
            except (KeyError, ZeroDivisionError):
                pass

            # Add to the main dictionary
            if len(data) > 1:
                if s_name in self.featurecounts_data:
                    log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                self.add_data_source(f, s_name)
                self.featurecounts_data[s_name] = data

    def featurecounts_stats_table(self):
        """Take the parsed stats from the featureCounts report and add them to the
        basic stats table at the top of the report"""

        headers = {
            "percent_assigned": {
                "title": "% Assigned",
                "description": "% Assigned reads",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "RdYlGn",
            },
            "Assigned": {
                "title": f"{config.read_count_prefix} Assigned",
                "description": f"Assigned reads ({config.read_count_desc})",
                "min": 0,
                "scale": "PuBu",
                "modify": lambda x: float(x) * config.read_count_multiplier,
                "shared_key": "read_count",
            },
        }
        self.general_stats_addcols(self.featurecounts_data, headers)

    def featureCounts_chart(self):
        """Make the featureCounts assignment rates plot"""

        headers = {}
        for h in self.featurecounts_keys:
            nice_name = h.replace("Unassigned_", "Unassigned: ").replace("_", " ")
            nice_name = re.sub(r"([a-z])([A-Z])", r"\g<1> \g<2>", nice_name)
            headers[h] = {"name": nice_name}

        # Config for the plot
        config = {
            "id": "featureCounts_assignment_plot",
            "title": "featureCounts: Assignments",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
        }

        return bargraph.plot(self.featurecounts_data, headers, config)



import logging

from ..base_module import ModuleNoSamplesFound
from .coverage_hist import DragenCoverageHist
from .coverage_metrics import DragenCoverageMetrics
from .coverage_per_contig import DragenCoveragePerContig
from .dragen_gc_metrics import DragenGcMetrics
from .fragment_length import DragenFragmentLength
from .mapping_metrics import DragenMappingMetics
from .overall_mean_cov import DragenOverallMeanCovMetrics
from .ploidy_estimation_metrics import DragenPloidyEstimationMetrics
from .rna_quant_metrics import DragenRnaQuantMetrics
from .rna_transcript_cov import DragenRnaTranscriptCoverage
from .sc_atac_metrics import DragenScAtacMetrics
from .sc_rna_metrics import DragenScRnaMetrics
from .time_metrics import DragenTimeMetrics
from .trimmer_metrics import DragenTrimmerMetrics
from .vc_metrics import DragenVCMetrics

log = logging.getLogger(__name__)


class MultiqcModule(
    DragenMappingMetics,
    DragenFragmentLength,
    DragenPloidyEstimationMetrics,
    DragenVCMetrics,
    DragenCoveragePerContig,
    DragenOverallMeanCovMetrics,
    DragenCoverageMetrics,
    DragenCoverageHist,
    DragenGcMetrics,
    DragenTrimmerMetrics,
    DragenTimeMetrics,
    DragenRnaQuantMetrics,
    DragenRnaTranscriptCoverage,
    DragenScRnaMetrics,
    DragenScAtacMetrics,
):
    """DRAGEN provides a number of different pipelines and outputs, including base calling, DNA and RNA alignment,
    post-alignment processing and variant calling, covering virtually all stages of typical NGS data processing.
    However, it can be treated as a fast aligner with additional features on top, as users will unlikely use any
    features without enabling DRAGEN mapping. So we will treat this module as an alignment tool module and
    place it accordingly in the module_order list, in docs, etc.

    The QC metrics DRAGEN generates resemble those of samtools-stats, qualimap, mosdepth, bcftools-stats and alike.
    Whenever possible, the visual output is made similar to those modules.

    Note that this MultiQC module supports some of DRAGEN output but not all. Contributions are welcome!

    The code is structured in a way so every mix-in parses one type of QC file that DRAGEN generates
    (e.g. *.mapping_metrics.csv, *.wgs_fine_hist_normal.csv, etc.). If a corresponding file is found, a mix-in adds
    a section into the report.
    """

    def __init__(self):
        super(MultiqcModule, self).__init__(
            name="DRAGEN",
            anchor="DRAGEN",
            target="DRAGEN",
            href="https://www.illumina.com/products/by-type/informatics-products/dragen-bio-it-platform.html",
            info=""" is a Bio-IT Platform that provides ultra-rapid secondary analysis of sequencing data
                     using field-programmable gate array technology (FPGA).""",
            # Can't find a DOI // doi=
        )

        samples_found = set()
        samples_found |= self.add_mapping_metrics()
        # <output prefix>.mapping_metrics.csv              - general stats table, a dedicated table, and a few barplots

        samples_found |= self.add_vc_metrics()
        # <output prefix>.vc_metrics.csv                   - a dedicated table and the total number of Variants into the general stats table

        samples_found |= self.add_ploidy_estimation_metrics()
        # <output prefix>.ploidy_estimation_metrics.csv    - add just Ploidy estimation into gen stats

        overall_mean_cov_data = self.collect_overall_mean_cov_data()
        # <output prefix>.<coverage region prefix>_overall_mean_cov<arbitrary suffix>.csv
        # This data will be used by in the DragenCoverageMetrics.

        samples_found |= self.add_coverage_metrics(overall_mean_cov_data)
        # <output prefix>.<coverage region prefix>_coverage_metrics<arbitrary suffix>.csv

        samples_found |= self.add_coverage_hist()

        # <output prefix>.wgs_fine_hist_normal.csv         - coverage distribution and cumulative coverage plots
        # <output prefix>.wgs_fine_hist_tumor.csv          - same

        samples_found |= self.add_coverage_per_contig()
        # <output prefix>.wgs_contig_mean_cov_normal.csv   - a histogram like in mosdepth, with each chrom as a category on X axis, plus a category for autosomal chromosomes average
        # <output prefix>.wgs_contig_mean_cov_tumor.csv    - same

        samples_found |= self.add_fragment_length_hist()
        # <output prefix>.fragment_length_hist.csv         - a histogram plot

        samples_found |= self.add_gc_metrics_hist()
        # <output prefix>.gc_metrics.csv

        samples_found |= self.add_trimmer_metrics()
        # <output prefix>.trimmer_metrics.csv

        samples_found |= self.add_time_metrics()
        # <output prefix>.time_metrics.csv

        samples_found |= self.add_rna_metrics()
        # <output prefix>.quant.metrics.csv

        samples_found |= self.add_rna_transcript_coverage()
        # <output prefix>.quant.transcript_coverage.txt

        samples_found |= self.add_sc_rna_metrics()
        # <output prefix>.scRNA.metrics.csv or <output prefix>.scRNA_metrics.csv

        samples_found |= self.add_sc_atac_metrics()
        # <output prefix>.scATAC.metrics.csv or <output prefix>.scATAC_metrics.csv

        if len(samples_found) == 0:
            raise ModuleNoSamplesFound
        log.info(f"Found samples: {len(samples_found)}")



""" MultiQC submodule to parse output from deepTools plotFingerprint """

import logging

import numpy as np

from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


class PlotFingerprintMixin:
    def parse_plotFingerprint(self):
        """Find plotFingerprint output. Both --outQualityMetrics and --outRawCounts"""
        self.deeptools_plotFingerprintOutQualityMetrics = dict()
        for f in self.find_log_files("deeptools/plotFingerprintOutQualityMetrics"):
            parsed_data = self.parsePlotFingerprintOutQualityMetrics(f)
            for k, v in parsed_data.items():
                if k in self.deeptools_plotFingerprintOutQualityMetrics:
                    log.warning(f"Replacing duplicate sample {k}.")
                # Values are fractions - convert to percentages for consistency with other MultiQC output
                self.deeptools_plotFingerprintOutQualityMetrics[k] = {i: float(j) * 100.0 for i, j in v.items()}

            if len(parsed_data) > 0:
                self.add_data_source(f, section="plotFingerprint")

        self.deeptools_plotFingerprintOutRawCounts = dict()
        for f in self.find_log_files("deeptools/plotFingerprintOutRawCounts"):
            parsed_data = self.parsePlotFingerprintOutRawCounts(f)
            for k, v in parsed_data.items():
                if k in self.deeptools_plotFingerprintOutRawCounts:
                    log.warning(f"Replacing duplicate sample {k}.")
                self.deeptools_plotFingerprintOutRawCounts[k] = v

            if len(parsed_data) > 0:
                self.add_data_source(f, section="plotFingerprint")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        self.deeptools_plotFingerprintOutRawCounts = self.ignore_samples(self.deeptools_plotFingerprintOutRawCounts)
        self.deeptools_plotFingerprintOutQualityMetrics = self.ignore_samples(
            self.deeptools_plotFingerprintOutQualityMetrics
        )

        if len(self.deeptools_plotFingerprintOutRawCounts) > 0:
            # Write data to file
            self.write_data_file(self.deeptools_plotFingerprintOutRawCounts, "deeptools_plot_fingerprint_counts")

            self.add_section(
                name="Fingerprint plot",
                anchor="deeptools_fingerprint",
                description="Signal fingerprint according to plotFingerprint",
                plot=linegraph.plot(
                    self.deeptools_plotFingerprintOutRawCounts,
                    {
                        "id": "deeptools_fingerprint_plot",
                        "title": "deepTools: Fingerprint plot",
                        "xmin": 0.0,
                        "xmax": 1.0,
                        "ymin": 0.0,
                        "ymax": 1.0,
                        "xlab": "rank",
                        "ylab": "Fraction w.r.t. bin with highest coverage",
                    },
                ),
            )

        if len(self.deeptools_plotFingerprintOutQualityMetrics) > 0:
            # Write data to file
            self.write_data_file(self.deeptools_plotFingerprintOutQualityMetrics, "deeptools_plot_fingerprint_metrics")

            self.add_section(
                name="Fingerprint quality metrics",
                anchor="plotFingerprint",
                description="Various quality metrics returned by plotFingerprint",
                plot=linegraph.plot(
                    self.deeptools_plotFingerprintOutQualityMetrics,
                    {
                        "id": "plotFingerprint_quality_metrics",
                        "title": "deepTools: Fingerprint quality metrics",
                        "stacking": None,
                        "ymin": 0,
                        "ymax": 100,
                        "yLabelFormat": "{value}%",
                        "ylab": "Percentage of fragments",
                        "categories": True,
                        "tt_label": "<strong>{point.x}</strong>: {point.y:.2f}%",
                    },
                ),
            )

        return len(self.deeptools_plotFingerprintOutQualityMetrics), len(self.deeptools_plotFingerprintOutRawCounts)

    def parsePlotFingerprintOutQualityMetrics(self, f):
        d = {}
        firstLine = True
        header = []
        for line in f["f"].splitlines():
            cols = line.strip().split("\t")

            if len(cols) < 7:
                log.warning(
                    "{} was initially flagged as the output from plotFingerprint --outQualityMetrics, but that seems to not be the case. Skipping...".format(
                        f["fn"]
                    )
                )
                return dict()

            if firstLine:
                header = [str(x) for x in cols[1:]]
                firstLine = False
                continue

            s_name = self.clean_s_name(cols[0], f)
            if s_name in d:
                log.warning(f"Replacing duplicate sample {s_name}.")
            d[s_name] = dict()

            try:
                for i, c in enumerate(cols[1:]):
                    if i >= len(header):
                        log.warning(
                            "{} was initially flagged as the output from plotFingerprint --outQualityMetrics, but that seems to not be the case. Skipping...".format(
                                f["fn"]
                            )
                        )
                        return dict()
                    if header[i] == "AUC" or header[i] == "Synthetic AUC":
                        continue
                    d[s_name][header[i]] = float(c)
            except:  # noqa: E722
                log.warning(
                    "{} was initially flagged as the output from plotFingerprint --outQualityMetrics, but that seems to not be the case. Skipping...".format(
                        f["fn"]
                    )
                )
                return dict()
        return d

    def parsePlotFingerprintOutRawCounts(self, f):
        d = dict()
        samples = []
        firstLine = True
        for line in f["f"].splitlines():
            cols = line.strip().split("\t")
            if cols[0] == "#plotFingerprint --outRawCounts":
                continue

            if firstLine:
                for c in cols:
                    c = str(c).strip("'")
                    s_name = self.clean_s_name(c, f)
                    d[s_name] = []
                    samples.append(s_name)
                firstLine = False
                continue

            for idx, c in enumerate(cols):
                d[samples[idx]].append(self._int(c))

        # Switch to numpy, get the normalized cumsum
        x = np.linspace(
            0, len(d[samples[0]]) - 1, endpoint=True, num=100, dtype=int
        )  # The indices into the vectors that we'll actually return for plotting
        xp = np.arange(len(d[samples[0]]) + 1) / float(len(d[samples[0]]) + 1)
        for k, v in d.items():
            v = np.array(v)
            v = np.sort(v)
            cs = np.cumsum(v)
            cs = cs / float(cs[-1])
            # Convert for plotting
            v2 = dict()
            v2[0.0] = 0.0
            for _ in x:
                v2[xp[_]] = cs[_]
            d[k] = v2
        return d



"""MultiQC module to parse output from OUS variant calling pipeline"""


import csv
import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import table

log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        """MultiQC module for processing hap.py output logs"""
        super(MultiqcModule, self).__init__(
            name="hap.py",
            anchor="happy",
            href="https://github.com/Illumina/hap.py",
            info="is a set of programs based on htslib to benchmark variant calls against gold standard truth datasets.",
            # No publication / DOI // doi=
        )

        self.happy_raw_sample_names = set()
        self.happy_indel_data = dict()
        self.happy_snp_data = dict()

        for f in self.find_log_files("happy", filehandles=True):
            self.parse_file(f)
            self.add_data_source(f)

        if len(self.happy_raw_sample_names) == 0:
            raise ModuleNoSamplesFound

        # print number of happy reports found and parsed
        log.info(f"Found {len(self.happy_raw_sample_names)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to file
        self.write_data_file(self.happy_indel_data, "multiqc_happy_indel_data", data_format="json")
        self.write_data_file(self.happy_snp_data, "multiqc_happy_snp_data", data_format="json")

        # add sections with the values from the indel and snp happy output
        self.add_section(
            name="INDEL",
            anchor="happy-indel-plot",
            description="The default shown fields should give the best overview of quality, but there are many other hidden fields available.",
            helptext="""
                No plots are generated, as hap.py is generally run on single control samples (NA12878, etc.)

                Ideally, precision, recall and F1 Score should all be as close to 1 as possible.
            """,
            plot=table.plot(
                self.happy_indel_data,
                self.gen_headers("_indel"),
                pconfig={
                    "id": "happy_indel_plot",
                    "title": "Happy: INDEL",
                },
            ),
        )

        self.add_section(
            name="SNP",
            anchor="happy-snp-plot",
            description="The default shown fields should give the best overview of quality, but there are many other hidden fields available.",
            helptext="""
                No plots are generated, as hap.py is generally run on single control samples (NA12878, etc.)

                Ideally, precision, recall and F1 Score should all be as close to 1 as possible.
            """,
            plot=table.plot(
                self.happy_snp_data,
                self.gen_headers("_snp"),
                pconfig={"id": "happy_snp_plot", "title": "Happy: SNP"},
            ),
        )

    def parse_file(self, f):
        # Check that we're not ignoring this sample name
        if self.is_ignore_sample(f["s_name"]):
            return

        if f["s_name"] in self.happy_raw_sample_names:
            log.warning(f"Duplicate sample name found in {f['root']}! Overwriting: {f['s_name']}")
        self.happy_raw_sample_names.add(f["s_name"])

        rdr: csv.DictReader = csv.DictReader(f["f"])
        for row in rdr:
            row_id = f"{f['s_name']}_{row['Type']}_{row['Filter']}"
            if row["Type"] == "INDEL":
                if row_id not in self.happy_indel_data:
                    self.happy_indel_data[row_id] = {"sample_id": f["s_name"]}
                # add suffix for headers to differentiate between indel and snp
                for fn in rdr.fieldnames:
                    self.happy_indel_data[row_id][fn + "_indel"] = row[fn]

            elif row["Type"] == "SNP":
                if row_id not in self.happy_snp_data:
                    self.happy_snp_data[row_id] = {"sample_id": f["s_name"]}
                for fn in rdr.fieldnames:
                    self.happy_snp_data[row_id][fn + "_snp"] = row[fn]

    def gen_headers(self, suffix=""):
        h = {
            "METRIC.Recall": {  # string must match headers in the input file
                "title": "Recall",  # whatever string to be displayed in the html report table
                "description": "Recall for truth variant representation = TRUTH.TP / (TRUTH.TP + TRUTH.FN)",
                "min": 0,
                "max": 1,
                "format": "{:.4f}",
            },
            "METRIC.Precision": {
                "title": "Precision",
                "description": "Precision of query variants = QUERY.TP / (QUERY.TP + QUERY.FP)",
                "min": 0,
                "max": 1,
                "format": "{:.4f}",
            },
            "METRIC.Frac_NA": {
                "title": "Fraction NA",
                "description": "Fraction of non-assessed query calls = QUERY.UNK / QUERY.TOTAL",
                "min": 0,
                "max": 1,
                "format": "{:.4f}",
            },
            "METRIC.F1_Score": {
                "title": "F1 Score",
                "description": "Harmonic mean of precision and recall = 2METRIC.RecallMetric.Precision/(METRIC.Recall + METRIC.Precision)",
                "min": 0,
                "max": 1,
                "format": "{:.4f}",
            },
            "TRUTH.TOTAL": {
                "title": "Truth: Total",
                "description": "Total number of truth variants",
                "format": None,
                "hidden": True,
            },
            "TRUTH.TP": {
                "title": "Truth: True Positive",
                "description": "Number of true-positive calls in truth representation (counted via the truth sample column)",
                "format": None,
                "hidden": True,
            },
            "TRUTH.FN": {
                "title": "Truth: False Negative",
                "description": "Number of false-negative calls = calls in truth without matching query call",
                "format": None,
                "hidden": True,
            },
            "QUERY.TOTAL": {
                "title": "Query: Total",
                "description": "Total number of query calls",
                "format": None,
                "hidden": True,
            },
            "QUERY.TP": {
                "title": "Query: True Positive",
                "description": "Number of true positive calls in query representation (counted via the query sample column)",
                "format": None,
                "hidden": True,
            },
            "QUERY.FP": {
                "title": "Query: False Positive",
                "description": "Number of false-positive calls in the query file (mismatched query calls within the confident regions)",
                "format": None,
                "hidden": True,
            },
            "QUERY.UNK": {
                "title": "Query: Unknown",
                "description": "Number of query calls outside the confident regions",
                "format": None,
                "hidden": True,
            },
            "FP.gt": {
                "title": "False Positive genotype",
                "description": "Number of genotype mismatches (alleles match, but different zygosity)",
                "hidden": True,
            },
            "FP.al": {
                "title": "False Positive allele",
                "description": "Number of allele mismatches (variants matched by position and not by haplotype)",
                "hidden": True,
            },
            "TRUTH.TOTAL.TiTv_ratio": {
                "title": "Truth: Total TiTv ratio",
                "description": "Transition / Transversion ratio for all truth variants",
                "hidden": True,
                "format": "{:.4f}",
            },
            "TRUTH.TOTAL.het_hom_ratio": {
                "title": "Truth: Total het/hom ratio",
                "description": "Het/Hom ratio for all truth variants",
                "hidden": True,
                "format": "{:.4f}",
            },
            "TRUTH.FN.TiTv_ratio": {
                "title": "Truth: False Negative TiTv ratio",
                "description": "Transition / Transversion ratio for false-negative variants",
                "hidden": True,
                "format": "{:.4f}",
            },
            "TRUTH.FN.het_hom_ratio": {
                "title": "Truth: False Negative het/hom ratio",
                "description": "Het/Hom ratio for false-negative variants",
                "hidden": True,
                "format": "{:.4f}",
            },
            "TRUTH.TP.TiTv_ratio": {
                "title": "Truth: True Positive TiTv ratio",
                "description": "Transition / Transversion ratio for true positive variants",
                "hidden": True,
                "format": "{:.4f}",
            },
            "TRUTH.TP.het_hom_ratio": {
                "title": "Truth: True Positive het/hom ratio",
                "description": "Het/Hom ratio for true positive variants",
                "hidden": True,
                "format": "{:.4f}",
            },
            "QUERY.FP.TiTv_ratio": {
                "title": "Query: False Positive TiTv ratio",
                "description": "Transition / Transversion ratio for false positive variants",
                "hidden": True,
                "format": "{:.4f}",
            },
            "QUERY.FP.het_hom_ratio": {
                "title": "Query: False Positive het/hom ratio",
                "description": "Het/Hom ratio for false-positive variants",
                "hidden": True,
                "format": "{:.4f}",
            },
            "QUERY.TOTAL.TiTv_ratio": {
                "title": "Query: total TiTv ratio",
                "description": "Transition / Transversion ratio for all query variants",
                "hidden": True,
                "format": "{:.4f}",
            },
            "QUERY.TOTAL.het_hom_ratio": {
                "title": "Query: Total het/hom ratio",
                "description": "Het/Hom ratio for all query variants",
                "hidden": True,
                "format": "{:.4f}",
            },
            "QUERY.TP.TiTv_ratio": {
                "title": "Query: True Positive TiTv ratio",
                "description": "Transition / Transversion ratio for true positive variants (query representation)",
                "hidden": True,
                "format": "{:.4f}",
            },
            "QUERY.TP.het_hom_ratio": {
                "title": "Query: True Positive het/hom ratio",
                "description": "Het/Hom ratio for true positive variants (query representation)",
                "hidden": True,
                "format": "{:.4f}",
            },
            "QUERY.UNK.TiTv_ratio": {
                "title": "Query: Uknown TiTv ratio",
                "description": "Transition / Transversion ratio for unknown variants",
                "hidden": True,
                "format": "{:.4f}",
            },
            "QUERY.UNK.het_hom_ratio": {
                "title": "Query: Unknown het/hom ratio",
                "description": "Het/Hom ratio for unknown variants",
                "hidden": True,
                "format": "{:.4f}",
            },
            "Subset.Size": {
                "title": "Subset Size",
                "description": "the number of nucleotides contained in the current subset",
                "format": None,
                "hidden": True,
            },
            "Subset.IS_CONF.Size": {
                "title": "Subset Confident Size",
                "description": "This gives the number of confident bases (-f regions) in the current subset",
                "format": None,
                "hidden": True,
            },
        }

        # rename column headers with '_indel' or '_snp' suffix
        headers = [k + suffix for k in h.keys()]
        # recreate the ordered dictionary with all headers and information
        return dict(zip(headers, h.values()))



""" MultiQC module to parse output from RSeQC """

import logging
import os

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """RSeQC is a collection of scripts. This MultiQC module
    supports some but not all. The code for each script is split
    into its own file and adds a section to the module output if
    logs are found."""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="RSeQC",
            anchor="rseqc",
            href="http://rseqc.sourceforge.net/",
            info="package provides a number of useful modules that can"
            " comprehensively evaluate high throughput RNA-seq data.",
            doi="10.1093/bioinformatics/bts356",
        )

        # Set up class objects to hold parsed data
        self.general_stats_headers = dict()
        self.general_stats_data = dict()
        n = dict()

        # Get the list of submodules (can be customised)
        rseqc_sections = getattr(config, "rseqc_sections", [])
        if len(rseqc_sections) == 0:
            rseqc_sections = [
                "read_distribution",
                "gene_body_coverage",
                "inner_distance",
                "read_gc",
                "read_duplication",
                "junction_annotation",
                "junction_saturation",
                "infer_experiment",
                "bam_stat",
                "tin",
            ]

        # Add self.js to be included in template
        self.js = {
            "assets/js/multiqc_rseqc.js": os.path.join(os.path.dirname(__file__), "assets", "js", "multiqc_rseqc.js")
        }

        # Call submodule functions
        for sm in rseqc_sections:
            try:
                # Import the submodule and call parse_reports()
                #   Function returns number of parsed logs
                module = __import__(f"multiqc.modules.rseqc.{sm}", fromlist=[""])
                n[sm] = getattr(module, "parse_reports")(self)
                if n[sm] > 0:
                    log.info(f"Found {n[sm]} {sm} reports")
            except (ImportError, AttributeError):
                log.error(f"Could not find RSeQC Section '{sm}'")

        # Exit if we didn't find anything
        if sum(n.values()) == 0:
            raise ModuleNoSamplesFound

        # Add to the General Stats table (has to be called once per MultiQC module)
        if max((len(vals) for vals in self.general_stats_data.values()), default=0) > 0:
            self.general_stats_addcols(self.general_stats_data, self.general_stats_headers)



'''"""""""""""""""""""""""""""""""""""""""""""""""""""""""
This module gathers data from _overall_mean_cov.csv files.
It relies on the following official source:
Overall Mean Coverage Report, page 196.
https://emea.support.illumina.com/content/dam/illumina-support/documents/documentation/software_documentation/dragen-bio-it/Illumina-DRAGEN-Bio-IT-Platform-User-Guide-1000000141465-00.pdf
'''

import logging
import re
from collections import defaultdict

from multiqc.modules.base_module import BaseMultiqcModule

from .utils import make_log_report

# Initialise the logger.
log = logging.getLogger(__name__)


class DragenOverallMeanCovMetrics(BaseMultiqcModule):
    """Public members of the DragenOverallMeanCovMetrics module are available to
    other dragen modules. Defines 2 public functions and a private data holder.
    """

    def collect_overall_mean_cov_data(self):
        """Collects raw data for coverage_metrics.py from overall_mean_cov.csv files."""
        overall_mean_cov_data = defaultdict(lambda: defaultdict(dict))

        for file in self.find_log_files("dragen/overall_mean_cov_metrics"):
            out = parse_overall_mean_cov(file)

            if out["success"]:
                self.add_data_source(file, section="overall_mean_cov_metrics")

                # Data for the coverage_metrics.py module.
                if out["phenotype"]:
                    sample, phenotype, data = out["sample"], out["phenotype"], out["data"]
                    overall_mean_cov_data[file["root"]][sample][phenotype] = data
                # Currently there is no need to support other files. Pass for now.
                else:
                    pass

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, file["s_name"])

        # Report found info/warnings/errors, which were collected while
        # calling the coverage_parser and constructing cov_headers.
        # You can disable it anytime, if it is not wanted.
        make_log_report(log_data, log, "overall_mean_cov_metrics")

        # No need to write the data.
        self.write_data_file(overall_mean_cov_data, "dragen_overall_mean_cov_data")

        # Just to pass the pre-commit
        # doi=

        return overall_mean_cov_data


'''"""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""""
The log_data is the container for found info, warnings and errors.
Collected data is stored in groups to make the output log file more readable.

warnings:
- invalid_file_names, which do not conform to the:
    <output-prefix>_overall_mean_cov<arbitrary-suffix>.csv

debug/info:
- unknown_metrics are those, which could not be recognized by the code.
  Such metrics will be ignored.
- unusual_values are those except for float.
'''
log_data = {
    "invalid_file_names": defaultdict(list),
    "unknown_metrics": [],
    "unusual_values": defaultdict(lambda: defaultdict(dict)),
}

# Official structure of files:   _overall_mean_cov.csv
# Accepted structure of files: .+_overall_mean_cov.*.csv
GEN_FILE_RGX = re.compile(r"(.+)_overall_mean_cov(.*)\.csv$")

# Special case. Coverage metrics files have the following structure:
# <output-prefix>.<coverage-region-prefix>_overall_mean_cov<arbitrary-suffix>.csv
COV_FILE_RGX = re.compile(r"(.+)\.(.+)_overall_mean_cov(.*)\.csv$")

# General structure of lines is not defined.
# Currently only 1 metric is present in the standard. It substitutes the line's regex.
AVG_RGX = re.compile(r"Average alignment coverage over ([^,]+),([^,]+)$", re.IGNORECASE)


def parse_overall_mean_cov(file_handler):
    """Parser for _overall_mean_cov.csv files."""

    root = file_handler["root"]
    file = file_handler["fn"]

    # First check the general structure.
    file_match = GEN_FILE_RGX.search(file)
    if file_match:
        sample, phenotype = file_match.group(1), None
        file_match = COV_FILE_RGX.search(file)
        if file_match:
            sample, phenotype, suffix = file_match.groups()
            if suffix:
                # Only non-'tumor' strings are concatenated with the sample.
                if not re.search("tumor", suffix, re.IGNORECASE):
                    sample += suffix
    else:
        log_data["invalid_file_names"][root].append(file)
        return {"success": False}

    # There is still a possibility for failing(eg empty file, unknown metric).
    success = False
    source = None  # File name from "Average alignment coverage over file" metric.
    value = None

    # A loop is created in advance. Maybe new metrics will be added in the future.
    for line in file_handler["f"].splitlines():
        line_match = AVG_RGX.search(line)
        # If line is fine then extract the necessary fields.
        if line_match:
            success = True
            source, value = line_match.groups()

        # Otherwise check if line is empty. If not then report it and go to the next line.
        else:
            if not re.search(r"^\s*$", line):
                log_data["unknown_metrics"].append(line)
            continue

        # Try to convert the value. It shall be float.
        try:
            value = float(value)
        except ValueError:
            log_data["unusual_values"][root][file][line] = value

    return {
        "success": success,
        "sample": sample,
        "phenotype": phenotype,
        "data": {"source_file": source, "value": value},
    }



""" MultiQC submodule to parse output from Picard RrbsSummaryMetrics """

import logging
from collections import OrderedDict

from multiqc.modules.picard import util
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(module):
    """Find Picard RrbsSummaryMetrics reports and parse their data"""

    data_by_sample = dict()

    # Go through logs and find Metrics
    for f in module.find_log_files("picard/rrbs_metrics", filehandles=True):
        s_name = None
        for line in f["f"]:
            maybe_s_name = util.extract_sample_name(
                module,
                line,
                f,
                picard_tool="CollectRrbsMetrics",
            )
            if maybe_s_name:
                s_name = maybe_s_name
                continue

            if s_name is None:
                continue

            if util.is_line_right_before_table(line, picard_class="RrbsSummaryMetrics"):
                if s_name in data_by_sample:
                    log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")
                module.add_data_source(f, s_name, section="RnaSeqMetrics")
                data_by_sample[s_name] = dict()

                keys = f["f"].readline().strip("\n").split("\t")
                vals = f["f"].readline().strip("\n").split("\t")
                if len(vals) != len(keys):
                    continue

                for k, v in zip(keys, vals):
                    if not v:
                        v = "NA"
                    else:
                        try:
                            v = float(v)
                        except ValueError:
                            pass
                    data_by_sample[s_name][k] = v

    # Filter to strip out ignored sample names
    data_by_sample = module.ignore_samples(data_by_sample)
    if len(data_by_sample) == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    module.add_software_version(None)

    # Write parsed data to a file
    module.write_data_file(data_by_sample, "multiqc_picard_RrbsSummaryMetrics")

    # Add to general stats table
    headers = {
        "PCT_CPG_BASES_CONVERTED": {
            "title": "CpG Methylated",
            "description": "Percentage of times a CpG cytosine was converted",
            "max": 100,
            "min": 0,
            "suffix": "%",
            "format": "{:,.0f}",
            "scale": "RdYlGn-rev",
            "modify": lambda x: 100 - util.multiply_hundred(x),
        },
        "PCT_NON_CPG_BASES_CONVERTED": {
            "title": "Non-CpG Methylated",
            "description": "Percentage of times a non-CpG cytosine was converted",
            "max": 100,
            "min": 0,
            "suffix": "%",
            "format": "{:,.0f}",
            "scale": "RdYlGn",
            "modify": lambda x: 100 - util.multiply_hundred(x),
        },
        "MEDIAN_CPG_COVERAGE": {
            "title": "Median CpG Cov",
            "description": "Median coverage of CpG sites",
            "min": 0,
        },
    }
    module.general_stats_addcols(data_by_sample, headers, namespace="RrbsSummaryMetrics")

    # Make the bar plot of converted bases
    pdata_cpg = dict()
    pdata_noncpg = dict()
    for s_name in data_by_sample.keys():
        pdata_cpg[s_name] = dict()
        pdata_cpg[s_name]["converted"] = data_by_sample[s_name]["CPG_BASES_CONVERTED"]
        pdata_cpg[s_name]["not_converted"] = (
            data_by_sample[s_name]["CPG_BASES_SEEN"] - data_by_sample[s_name]["CPG_BASES_CONVERTED"]
        )
        pdata_noncpg[s_name] = dict()
        pdata_noncpg[s_name]["converted"] = data_by_sample[s_name]["NON_CPG_BASES"]
        pdata_noncpg[s_name]["not_converted"] = (
            data_by_sample[s_name]["NON_CPG_BASES"] - data_by_sample[s_name]["NON_CPG_CONVERTED_BASES"]
        )

    keys = dict()
    keys["not_converted"] = {"name": "Unconverted Bases (Methylated)"}
    keys["converted"] = {"name": "Converted Bases (Unmethylated)"}

    # Config for the plot
    pconfig = {
        "id": "picard_rrbs_converted_bases_plot",
        "title": "Picard: Converted Bases",
        "ylab": "# CpG Bases",
        "cpswitch_counts_label": "Number of Bases",
        "data_labels": [{"name": "CpG", "ylab": "# CpG Bases"}, {"name": "Non-CpG", "ylab": "# Non-CpG Bases"}],
    }

    module.add_section(
        name="RRBS Converted Bases",
        anchor="picard-rrbssummary-convertedbases",
        plot=bargraph.plot([pdata_cpg, pdata_noncpg], [keys, keys], pconfig),
    )

    # Make the bar plot of processed reads
    pdata = dict()
    for s_name in data_by_sample:
        pdata[s_name] = dict()
        pdata[s_name]["with_no_cpg"] = data_by_sample[s_name]["READS_WITH_NO_CPG"]
        pdata[s_name]["ignored_short"] = data_by_sample[s_name]["READS_IGNORED_SHORT"]
        pdata[s_name]["ignored_mismatches"] = data_by_sample[s_name]["READS_IGNORED_MISMATCHES"]
        pdata[s_name]["not_ignored"] = (
            data_by_sample[s_name]["READS_ALIGNED"]
            - pdata[s_name]["with_no_cpg"]
            - pdata[s_name]["ignored_short"]
            - pdata[s_name]["ignored_mismatches"]
        )

    keys = OrderedDict()
    keys["not_ignored"] = {"name": "Utilised reads"}
    keys["with_no_cpg"] = {"name": "Ignored (no CpG sites)"}
    keys["ignored_short"] = {"name": "Ignored (too short)"}
    keys["ignored_mismatches"] = {"name": "Ignored (exceeded mismatch threshold)"}

    # Config for the plot
    pconfig = {
        "id": "picard_rrbs_ignored_reads_plot",
        "title": "Picard: RRBS Read Counts",
        "ylab": "# Reads",
        "cpswitch_counts_label": "Number of Reads",
    }

    module.add_section(
        name="RRBS Read Counts", anchor="picard-rrbssummary-readcounts", plot=bargraph.plot(pdata, keys, pconfig)
    )

    # Return the number of detected samples to the parent module
    return len(data_by_sample)



""" MultiQC module to parse output from SortMeRNA """


import logging
import os
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="SortMeRNA",
            anchor="sortmerna",
            href="http://bioinfo.lifl.fr/RNA/sortmerna/",
            info="is a program for filtering, mapping and OTU-picking NGS reads in metatranscriptomic and metagenomic data.",
            doi="10.1093/bioinformatics/bts611",
        )

        # Parse logs
        self.sortmerna = dict()
        for f in self.find_log_files("sortmerna", filehandles=True):
            self.parse_sortmerna(f)
            self.add_data_source(f)

        # Filter to strip out ignored sample names
        self.sortmerna = self.ignore_samples(self.sortmerna)

        if len(self.sortmerna) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.sortmerna)} logs")

        self.write_data_file(self.sortmerna, "multiqc_sortmerna")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Get custom table header, default to 'rRNA'
        tab_header = getattr(config, "sortmerna", {}).get("seqname", "% rRNA")

        # Add rRNA rate to the general stats table
        headers = {
            "rRNA_pct": {
                "title": tab_header,
                "description": "Percentage of reads matched to a SortMeRNA database",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "OrRd",
            }
        }
        self.general_stats_addcols(self.sortmerna, headers)

        # Make barplot
        self.sortmerna_detailed_barplot()

    def parse_sortmerna(self, f):
        s_name = None
        post_results_start = False
        post_database_start = False
        db_number = 0
        err = False

        for line in f["f"]:
            if "Reads file" in line:
                parts = re.split(r"[:=]", line)
                s_name = self.clean_s_name(parts[-1], f)
                self.sortmerna[s_name] = dict()
            if "Results:" in line and not post_results_start:  # old versions
                post_results_start = True
            if "Total reads = " in line and not post_results_start:  # v4.2.0 onwards
                post_results_start = True
            if not post_results_start:
                continue
            if post_results_start and not post_database_start:
                if "Total reads =" in line:
                    m = re.search(r"\d+", line)
                    if m:
                        self.sortmerna[s_name]["total"] = int(m.group())
                    else:
                        err = True
                elif "Total reads passing" in line:
                    m = re.search(r"\d+", line)
                    if m:
                        self.sortmerna[s_name]["rRNA"] = int(m.group())
                        self.sortmerna[s_name]["rRNA_pct"] = (
                            float(self.sortmerna[s_name]["rRNA"]) / float(self.sortmerna[s_name]["total"]) * 100
                        )
                    else:
                        err = True
                elif "Total reads failing" in line:
                    m = re.search(r"\d+", line)
                    if m:
                        self.sortmerna[s_name]["non_rRNA"] = int(m.group())
                        self.sortmerna[s_name]["non_rRNA_pct"] = (
                            float(self.sortmerna[s_name]["non_rRNA"]) / float(self.sortmerna[s_name]["total"]) * 100
                        )
                    else:
                        err = True
            if post_database_start:
                # Stop when we hit empty lines
                if not line.strip():
                    break

                db_number = db_number + 1
                parts = re.split("\t+", line.strip())
                if len(parts) == 2:
                    db = os.path.splitext(os.path.basename(parts[0]))[0]
                    pct = float(parts[1].replace("%", ""))
                    count = int(self.sortmerna[s_name]["total"]) * (pct / 100.0)
                    self.sortmerna[s_name][db + "_pct"] = pct
                    self.sortmerna[s_name][db + "_count"] = count
                else:
                    err = True

            if "By database:" in line or "Coverage by database:" in line:
                post_database_start = True
        if err:
            log.warning("Error parsing data in: " + s_name)
            self.sortmerna.pop(s_name, "None")
        s_name = None

    def sortmerna_detailed_barplot(self):
        """Make the HighCharts HTML to plot the sortmerna rates"""

        # Specify the order of the different possible categories
        keys = {}
        metrics = set()
        for sample in self.sortmerna:
            for key in self.sortmerna[sample]:
                if key not in ["total", "rRNA", "non_rRNA"] and "_pct" not in key:
                    metrics.add(key)

        for key in metrics:
            keys[key] = {"name": key.replace("_count", "")}

        # Config for the plot
        pconfig = {"id": "sortmerna-detailed-plot", "title": "SortMeRNA: Hit Counts", "ylab": "Reads"}

        self.add_section(plot=bargraph.plot(self.sortmerna, sorted(keys), pconfig))



""" MultiQC submodule to parse output from deepTools estimateReadFiltering """

import logging

from multiqc.plots import table

# Initialise the logger
log = logging.getLogger(__name__)


class EstimateReadFilteringMixin:
    def parse_estimate_read_filtering(self):
        """Find estimateReadFiltering output. Only the output from --table is supported."""
        self.deeptools_estimateReadFiltering = dict()
        for f in self.find_log_files("deeptools/estimateReadFiltering"):
            parsed_data = self.parse_estimate_read_filtering_file(f)
            for k, v in parsed_data.items():
                if k in self.deeptools_estimateReadFiltering:
                    log.warning(f"Replacing duplicate sample {k}.")
                self.deeptools_estimateReadFiltering[k] = v

            if len(parsed_data) > 0:
                self.add_data_source(f, section="estimateReadFiltering")

        self.deeptools_estimateReadFiltering = self.ignore_samples(self.deeptools_estimateReadFiltering)

        if len(self.deeptools_estimateReadFiltering) == 0:
            return 0

        # Write data to file
        self.write_data_file(self.deeptools_estimateReadFiltering, "deeptools_read_filtering")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        header = {
            "M Entries": {"title": "M entries", "description": "Number of entries in the file (millions)"},
            "pct_Aligned": {
                "title": "% Aligned",
                "description": "Percent of aligned entries",
                "scale": "YlGn",
                "min": 0,
                "max": 100,
            },
            "pct_Filtered": {
                "title": "% Tot. Filtered",
                "description": "Percent of alignment that would be filtered for any reason.",
                "scale": "OrRd",
                "min": 0,
                "max": 100,
            },
            "pct_Blacklisted": {
                "title": "% Blacklisted",
                "description": "Percent of alignments falling (at least partially) inside a blacklisted region",
                "scale": "YlOrRd",
                "min": 0,
                "max": 100,
            },
            "pct_MAPQ": {
                "title": "% MAPQ",
                "description": "Percent of alignments having MAPQ scores below the specified threshold",
                "scale": "YlOrBn",
                "min": 0,
                "max": 100,
            },
            "pct_Missing_Flags": {
                "title": "% Missing Flags",
                "description": "Percent of alignments lacking at least on flag specified by --samFlagInclude",
                "scale": "PuRd",
                "min": 0,
                "max": 100,
            },
            "pct_Forbidden_Flags": {
                "title": "% Forbidden Flags",
                "description": "Percent of alignments having at least one flag specified by --samFlagExclude",
                "scale": "OrRd",
                "min": 0,
                "max": 100,
            },
            "pct_deepTools_Dupes": {
                "title": "% deepTools Dupes",
                "description": "Percent of alignments marked by deepTools as being duplicates",
                "scale": "PuRd",
                "min": 0,
                "max": 100,
            },
            "pct_Duplication": {
                "title": "% Duplication",
                "description": "Percent of alignments originally marked as being duplicates",
                "scale": "OrRd",
                "min": 0,
                "max": 100,
            },
            "pct_Singletons": {
                "title": "% Singletons",
                "description": "Percent of alignments that are singletons (i.e., paired-end reads where the mates don't align as a pair",
                "scale": "OrRd",
                "min": 0,
                "max": 100,
            },
            "pct_Strand_Filtered": {
                "title": "% Strand Filtered",
                "description": "Percent of alignments arising from the wrong strand",
                "scale": "OrRd",
                "min": 0,
                "max": 100,
            },
        }

        tdata = dict()
        for k, v in self.deeptools_estimateReadFiltering.items():
            tdata[k] = {
                "M Entries": v["total"] / 1000000.0,
                "pct_Aligned": 100.0 * v["mapped"] / float(v["total"]),
                "pct_Filtered": 100.0 * v["filtered"] / float(v["total"]),
                "pct_Blacklisted": 100.0 * v["blacklisted"] / float(v["total"]),
                "pct_Below_MAPQ": 100.0 * v["mapq"] / float(v["total"]),
                "pct_Missing_Flags": 100.0 * v["required flags"] / float(v["total"]),
                "pct_Forbidden_Flags": 100.0 * v["excluded flags"] / float(v["total"]),
                "pct_deepTools_Dupes": 100.0 * v["internal dupes"] / float(v["total"]),
                "pct_Duplication": 100.0 * v["dupes"] / float(v["total"]),
                "pct_Singletons": 100.0 * v["singletons"] / float(v["total"]),
                "pct_Strand_Filtered": 100.0 * v["strand"] / float(v["total"]),
            }

        config = {
            "namespace": "deepTools bamPEFragmentSize",
            "id": "deeptools_estimate_read_filtering_table",
            "title": "deepTools: Estimate read filtering",
        }
        self.add_section(
            name="Filtering metrics",
            anchor="estimateReadFiltering",
            description="Estimated percentages of alignments filtered independently for each setting in `estimateReadFiltering`",
            plot=table.plot(tdata, header, config),
        )

        return len(self.deeptools_estimateReadFiltering)

    def parse_estimate_read_filtering_file(self, f):
        d = {}
        firstLine = True
        for line in f["f"].splitlines():
            if firstLine:
                firstLine = False
                continue
            cols = line.strip().split("\t")

            if len(cols) != 12:
                # This is not really the output from estimateReadFiltering!
                log.warning(
                    "{} was initially flagged as the tabular output from estimateReadFiltering, but that seems to not be the case. Skipping...".format(
                        f["fn"]
                    )
                )
                return dict()

            s_name = self.clean_s_name(cols[0], f)
            if s_name in d:
                log.debug(f"Replacing duplicate sample {s_name}.")
            d[s_name] = dict()

            try:
                d[s_name]["total"] = self._int(cols[1])
                d[s_name]["mapped"] = self._int(cols[2])
                d[s_name]["blacklisted"] = self._int(cols[3])
                d[s_name]["filtered"] = float(cols[4])
                d[s_name]["mapq"] = float(cols[5])
                d[s_name]["required flags"] = float(cols[6])
                d[s_name]["excluded flags"] = float(cols[7])
                d[s_name]["internal dupes"] = float(cols[8])
                d[s_name]["dupes"] = float(cols[9])
                d[s_name]["singletons"] = float(cols[10])
                d[s_name]["strand"] = float(cols[11])
            except:  # noqa: E722
                # Obviously this isn't really the output from estimateReadFiltering
                log.warning(
                    "{} was initially flagged as the output from estimateReadFiltering, but that seems to not be the case. Skipping...".format(
                        f["fn"]
                    )
                )
                return dict()
        return d



#! /usr/bin/env python

""" MultiQC module to parse output from Pangolin """


import csv
import logging
from typing import Optional

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import table
from multiqc.utils import mqc_colour

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """Pangolin module"""

    def __init__(self):
        # Initialise the parent module
        super().__init__(
            name="Pangolin",
            anchor="pangolin",
            href="https://github.com/cov-lineages/pangolin",
            info="uses variant calls to assign SARS-CoV-2 genome sequences to global lineages.",
            doi="10.1093/ve/veab064",
        )

        # Find and parse the sample files
        self.pangolin_data = dict()
        self.lineage_colours = dict()
        for f in self.find_log_files("pangolin", filehandles=True):
            self.parse_pangolin_log(f)
            self.add_data_source(f)

        # Filter out parsed samples based on sample name
        self.pangolin_data = self.ignore_samples(self.pangolin_data)

        # Stop if we didn't find anything
        if len(self.pangolin_data) == 0:
            raise ModuleNoSamplesFound
        log.info(f"Found {len(self.pangolin_data)} samples")
        self.write_data_file(self.pangolin_data, "multiqc_pangolin")

        # Assign some lineage colours
        # First, remove blank / None
        self.lineage_colours.pop("", None)
        self.lineage_colours.pop("None", None)
        cols = mqc_colour.mqc_colour_scale("Dark2", 0, len(self.lineage_colours))
        for idx, k in enumerate(self.lineage_colours):
            self.lineage_colours[k] = cols.get_colour(idx)
        # Manually add back None as grey
        self.lineage_colours["None"] = "#EFEFEF"

        self.pangolin_general_stats_table()

        self.add_section(
            name="Run table",
            anchor="pangolin-run",
            description="Statistics gathered from the input pangolin files. Hover over the column headers for descriptions and click _Help_ for more in-depth documentation.",
            helptext="""
            This table shows some of the metrics parsed by Pangolin.
            Hover over the column headers to see a description of the contents. Longer help text for certain columns is shown below:

            * **Conflict**
                * In the pangoLEARN decision tree model, a given sequence gets assigned to the most likely category based on known diversity.
                  If a sequence can fit into more than one category, the conflict score will be greater than `0` and reflect the number of categories the sequence could fit into.
                  If the conflict score is `0`, this means that within the current decision tree there is only one category that the sequence could be assigned to.
            * **Ambiguity score**
                * This score is a function of the quantity of missing data in a sequence.
                  It represents the proportion of relevant sites in a sequence which were imputed to the reference values.
                  A score of `1` indicates that no sites were imputed, while a score of `0` indicates that more sites were imputed than were not imputed.
                  This score only includes sites which are used by the decision tree to classify a sequence.
            * **Scorpio conflict**
                * The conflict score is the proportion of defining variants which have the reference allele in the sequence.
                  Ambiguous/other non-ref/alt bases at each of the variant positions contribute only to the denominators of these scores.
            * **Note**
                * If any conflicts from the decision tree, this field will output the alternative assignments.
                  If the sequence failed QC this field will describe why.
                  If the sequence met the SNP thresholds for scorpio to call a constellation, it’ll describe the exact SNP counts of Alt, Ref and Amb (Alternative, reference and ambiguous) alleles for that call.
            """,
            plot=self.pangolin_table(),
        )

    def parse_pangolin_log(self, f):
        def add_version_not_none(version, sample, name):
            if version is not None:
                self.add_software_version(version, sample, name)

        for row in csv.DictReader(f["f"]):
            try:
                taxon_name = row["taxon"]
                row.pop("taxon")
                # Taxon names sometimes include backslashes.
                # MultiQC assumes that these are file path separators and cleans them away
                # Bit of a nasty hack is to just replace them here first.
                taxon_name = taxon_name.replace("/", "_")
                s_name = self.clean_s_name(taxon_name, f)
                if s_name in self.pangolin_data:
                    log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                # Avoid generic header ID that clashes with other modules
                if "qc_status" not in row:
                    row["qc_status"] = row.pop("status")
                if "qc_notes" in row:
                    row["qc_notes"] = _format_qc_notes(row["qc_notes"])

                self.pangolin_data[s_name] = row
                # Just save the lineage key for now - we will sort out the colours later
                self.lineage_colours[row["lineage"]] = None
                self.lineage_colours[row["scorpio_call"]] = None

                # Version info
                # Note: Excluded "version" field from software versions as this refers to
                # how the reference data was prepared. This info is still available
                # in the "Run table" table
                add_version_not_none(row.get("pangolin_version"), s_name, self.name)
                add_version_not_none(row.get("pango_version"), s_name, "Pango")
                add_version_not_none(row.get("pangoLEARN_version"), s_name, "PangoLEARN")
                add_version_not_none(row.get("scorpio_version"), s_name, "Scorpio")
                # constellation_version is sometimes "TRUE" or "FALSE" - ignore these
                constellation_version = row.get("constellation_version")
                if constellation_version not in {None, "TRUE", "FALSE"}:
                    self.add_software_version(constellation_version, s_name, "Constellations")
            except KeyError:
                log.debug(f"File '{f['fn']}' could not be parsed - no taxon field found.")

    def pangolin_general_stats_table(self):
        """Takes the parsed sample data and adds it to the general stats table"""

        headers = {
            "lineage": {
                "title": "Lineage",
                "description": "Lineage",
                "min": 0,
                "scale": "RdYlGn",
                "bgcols": self.lineage_colours,
            }
        }
        self.general_stats_addcols(self.pangolin_data, headers)

    def pangolin_table(self):
        """Creates the table of all data for the samples"""

        headers = {
            "lineage": {
                "title": "Lineage",
                "description": """
                The most likely lineage assigned to a given sequence based on the inference engine used
                and the SARS-CoV-2 diversity designated
                """,
                "scale": False,
                "bgcols": self.lineage_colours,
            },
            "conflict": {
                "title": "Conflict",
                "description": "Conflict between categories in decision tree",
                "min": 0,
                "max": 1,
                "scale": "RdBu-rev",
            },
            "ambiguity_score": {
                "title": "Ambiguity",
                "description": "Quantity of missing data in a sequence",
                "min": 0,
                "max": 1,
                "scale": "RdYlGn",
            },
            "scorpio_call": {
                "title": "S call",
                "description": "Scorpio: If a query is assigned a constellation by scorpio this call is output in this column",
                "scale": False,
                "bgcols": self.lineage_colours,
            },
            "scorpio_support": {
                "title": "S support",
                "description": "Scorpio: The proportion of defining variants which have the alternative allele in the sequence",
                "min": 0,
                "max": 1,
                "scale": "RdYlBu",
            },
            "scorpio_conflict": {
                "title": "S conflict",
                "description": "Scorpio: The proportion of defining variants which have the reference allele in the sequence",
                "min": 0,
                "max": 1,
                "scale": "RdYlGn-rev",
            },
            "version": {
                "title": "Version",
                "description": "A version number that represents both the pango-designation number and the inference engine used to assign the lineage",
                "scale": False,
                "hidden": True,
            },
            "pangolin_version": {
                "title": "Pangolin version",
                "description": "The version of pangolin software running",
                "scale": False,
                "hidden": True,
            },
            "scorpio_version": {
                "title": "Scorpio version",
                "description": "The version of the scorpio software installed",
                "scale": False,
                "hidden": True,
            },
            "constellation_version": {
                "title": "Constellations version",
                "description": "The version of Constellations that scorpio has used to curate the lineage assignment",
                "scale": False,
                "hidden": True,
            },
            "qc_status": {
                "title": "QC Status",
                "description": "Indicates whether the sequence passed the QC thresholds for minimum length and maximum N content",
                "scale": False,
                "modify": lambda x: "Pass" if x == "passed_qc" else x.capitalize(),
            },
            "qc_notes": {
                "title": "QC Note",
                "description": "Notes specific to the QC checks run on the sequences",
                "scale": False,
            },
            "note": {
                "title": "Note",
                "description": "Additional information from Pangolin",
                "scale": False,
            },
        }

        # Main table config
        table_config = {
            "namespace": "Pangolin",
            "id": "pangolin_run_table",
            "table_title": "Pangolin Run details",
        }

        return table.plot(self.pangolin_data, headers, table_config)


def _format_qc_notes(raw: str) -> Optional[str]:
    """
    Parses QC notes, they appear to come from:
    https://github.com/cov-lineages/pangolin/blob/361f49cbffbf26eb28bed2f4a4c0e7f3d5a054cc/pangolin/utils/preprocessing.py#L91-L97
    https://github.com/cov-lineages/pangolin/blob/361f49cbffbf26eb28bed2f4a4c0e7f3d5a054cc/pangolin/utils/preprocessing.py#L179
    """
    if raw.startswith("Ambiguous_content"):
        # e.g. Ambiguous_content:0.03
        split = raw.split(":")
        if len(split) != 2:
            logging.warning(f"Expected label of format 'Ambiguous_content:0.01', found: '{raw}'")
            return None
        proportion_n = float(split[1])
        percent_n = int(proportion_n * 100)
        return f"Ambiguous content: {percent_n}%"

    # Unrecognized notes, just return them, capitalized
    return raw.capitalize()



import logging
from collections import defaultdict

from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import bargraph

log = logging.getLogger(__name__)


class DragenTimeMetrics(BaseMultiqcModule):
    def add_time_metrics(self):
        data_by_sample = dict()

        for f in self.find_log_files("dragen/time_metrics"):
            data = parse_time_metrics_file(f)
            s_name = f["s_name"]
            if s_name in data_by_sample:
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            self.add_data_source(f, section="time_metrics")
            data_by_sample[s_name] = data

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, s_name)

        # Filter to strip out ignored sample names:
        data_by_sample = self.ignore_samples(data_by_sample)

        if not data_by_sample:
            return set()

        # Save data file
        self.write_data_file(data_by_sample, "dragen_time_metrics")

        self.add_section(
            name="Time Metrics",
            anchor="dragen-time-metrics",
            description="Time metrics for DRAGEN run.  Total run time is less than the sum of individual "
            "steps because of parallelization.",
            plot=bargraph.plot(
                [
                    {
                        sample: {
                            step_name: step_time / 60
                            for step_name, step_time in data.items()
                            if step_name == "Total runtime"
                        }
                        for sample, data in data_by_sample.items()
                    },
                    {
                        sample: {
                            step_name[5:]: step_time / 60
                            for step_name, step_time in data.items()
                            if step_name != "Total runtime"
                        }
                        for sample, data in data_by_sample.items()
                    },
                ],
                pconfig={
                    "id": "time_metrics_plot",
                    "title": "Dragen: Time Metrics",
                    "ylab": "Time (minutes)",
                    "cpswitch_counts_label": "Time (minutes)",
                    "data_labels": [
                        {
                            "name": "Total Runtime",
                            "ylab": "Time (minutes)",
                            "cpswitch_counts_label": "Time (minutes)",
                        },
                        {
                            "name": "Steps Breakdown",
                            "ylab": "Time (minutes)",
                            "cpswitch_counts_label": "Time (minutes)",
                        },
                    ],
                },
            ),
        )

        return data_by_sample.keys()


def parse_time_metrics_file(f):
    """
    sample.time_metrics.csv

    RUN TIME,,Time loading reference,00:01:31.289,91.29
    RUN TIME,,Time aligning reads,00:00:25.190,25.19
    RUN TIME,,Time duplicate marking,00:00:01.817,1.82
    RUN TIME,,Time sorting and marking duplicates,00:00:07.368,7.37
    RUN TIME,,Time DRAGStr calibration,00:00:07.069,7.07
    """
    data = defaultdict(dict)
    for line in f["f"].splitlines():
        tokens = line.split(",")
        analysis, _, metric, timestr, seconds = tokens

        try:
            seconds = float(seconds)
        except ValueError:
            pass
        data[metric] = seconds

    return data



""" MultiQC submodule to parse output from deepTools plotCoverage """

import logging

from multiqc.plots import linegraph, table

# Initialise the logger
log = logging.getLogger(__name__)


class plotCoverageMixin:
    def parse_plotCoverage(self):
        """Find plotCoverage output. Both stdout and --outRawCounts"""
        self.deeptools_plotCoverageStdout = dict()
        for f in self.find_log_files("deeptools/plotCoverageStdout"):
            parsed_data = self.parsePlotCoverageStdout(f)
            for k, v in parsed_data.items():
                if k in self.deeptools_plotCoverageStdout:
                    log.warning(f"Replacing duplicate sample {k}.")
                self.deeptools_plotCoverageStdout[k] = v

            if len(parsed_data) > 0:
                self.add_data_source(f, section="plotCoverage")

        self.deeptools_plotCoverageOutRawCounts = dict()
        for f in self.find_log_files("deeptools/plotCoverageOutRawCounts"):
            parsed_data = self.parsePlotCoverageOutRawCounts(f)
            for k, v in parsed_data.items():
                if k in self.deeptools_plotCoverageOutRawCounts:
                    log.warning(f"Replacing duplicate sample {k}.")
                self.deeptools_plotCoverageOutRawCounts[k] = v

            if len(parsed_data) > 0:
                self.add_data_source(f, section="plotCoverage")

        self.deeptools_plotCoverageStdout = self.ignore_samples(self.deeptools_plotCoverageStdout)
        self.deeptools_plotCoverageOutRawCounts = self.ignore_samples(self.deeptools_plotCoverageOutRawCounts)

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        if len(self.deeptools_plotCoverageStdout) > 0:
            # Write data to file
            self.write_data_file(self.deeptools_plotCoverageStdout, "deeptools_plot_cov_stdout")

            header = {
                "min": {"title": "Min", "description": "Minimum Coverage", "shared_key": "coverage"},
                "25%": {
                    "rid": "first_quartile",
                    "title": "1st Quartile",
                    "description": "First quartile coverage",
                    "shared_key": "coverage",
                },
                "50%": {
                    "rid": "median",
                    "title": "Median",
                    "description": "Median coverage (second quartile)",
                    "shared_key": "coverage",
                },
                "mean": {"title": "Mean", "description": "Mean coverage", "shared_key": "coverage"},
                "75%": {
                    "rid": "third_quartile",
                    "title": "3rd Quartile",
                    "description": "Third quartile coverage",
                    "shared_key": "coverage",
                },
                "max": {"title": "Max", "description": "Maximum coverage", "shared_key": "coverage"},
                "std": {
                    "title": "Std. Dev.",
                    "description": "Coverage standard deviation",
                    "shared_key": "coverage",
                },
            }
            config = {
                "namespace": "deepTools plotCoverage",
                "id": "deeptools_coverage_metrics_table",
                "title": "deepTools: Coverage metrics",
            }
            self.add_section(
                name="Coverage metrics",
                anchor="deeptools_coverage_metrics",
                plot=table.plot(self.deeptools_plotCoverageStdout, header, config),
            )

        if len(self.deeptools_plotCoverageOutRawCounts) > 0:
            # Write data to file
            self.write_data_file(self.deeptools_plotCoverageOutRawCounts, "deeptools_plot_cov_counts")

            config = {
                "id": "deeptools_coverage_metrics_plot",
                "title": "deepTools: Coverage distribution",
                "xlab": "Coverage",
                "ylab": "Fraction of bases sampled",
            }
            self.add_section(
                name="Coverage distribution",
                anchor="deeptools_coverage_distribution",
                description="The fraction of bases with a given number of read/fragment coverage",
                plot=linegraph.plot(self.deeptools_plotCoverageOutRawCounts, config),
            )

        return len(self.deeptools_plotCoverageStdout), len(self.deeptools_plotCoverageOutRawCounts)

    def parsePlotCoverageStdout(self, f):
        d = {}
        firstLine = True
        for line in f["f"].splitlines():
            if firstLine:
                firstLine = False
                continue
            cols = line.strip().split("\t")

            if len(cols) != 8:
                log.warning(
                    "{} was initially flagged as the standard output from plotCoverage, but that seems to not be the case. Skipping...".format(
                        f["fn"]
                    )
                )
                return dict()

            s_name = self.clean_s_name(cols[0], f)
            if s_name in d:
                log.warning(f"Replacing duplicate sample {s_name}.")
            d[s_name] = dict()

            try:
                d[s_name]["mean"] = float(cols[1])
                d[s_name]["std"] = float(cols[2])
                d[s_name]["min"] = float(cols[3])
                d[s_name]["25%"] = float(cols[4])
                d[s_name]["50%"] = float(cols[5])
                d[s_name]["75%"] = float(cols[6])
                d[s_name]["max"] = float(cols[7])
            except:  # noqa: E722
                log.warning(
                    "{} was initially flagged as the standard output from plotCoverage, but that seems to not be the case. Skipping...".format(
                        f["fn"]
                    )
                )
                return dict()
        return d

    def parsePlotCoverageOutRawCounts(self, f):
        samples = []
        d = {}
        nCols = 0
        nRows = 0
        for line in f["f"].splitlines():
            if line.startswith("#plotCoverage"):
                continue

            cols = line.strip().split("\t")
            if len(cols) < 4:
                log.warning(
                    "{} was initially flagged as the output from plotCoverage --outRawCounts, but that seems to not be the case. Skipping...".format(
                        f["fn"]
                    )
                )
                return dict()

            if cols[0] == "#'chr'":
                nCols = len(cols)
                for col in cols[3:]:
                    s_name = self.clean_s_name(col.strip("'"), f)
                    samples.append(s_name)
                    d[s_name] = dict()
                continue

            if len(cols) != nCols:
                log.warning(
                    "{} was initially flagged as the output from plotCoverage --outRawCounts, but that seems to not be the case. Skipping...".format(
                        f["fn"]
                    )
                )
                return dict()

            for i, v in enumerate(cols[3:]):
                v = float(v)
                if v not in d[samples[i]]:
                    d[samples[i]][v] = 0
                d[samples[i]][v] += 1
            nRows += 1

        # Convert values to a fraction
        nRows = float(nRows)
        for k, v in d.items():
            for k2, v2 in v.items():
                v[k2] /= nRows

        return d



from typing import Dict

import logging
from collections import defaultdict

from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import table

log = logging.getLogger(__name__)


class DragenTrimmerMetrics(BaseMultiqcModule):
    def add_trimmer_metrics(self):
        data_by_sample = dict()

        for f in self.find_log_files("dragen/trimmer_metrics"):
            data = parse_trimmer_metrics_file(f)
            s_name = f["s_name"]
            if s_name in data_by_sample:
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            self.add_data_source(f, section="trimmer_metrics")
            data_by_sample[s_name] = data

        # Filter to strip out ignored sample names:
        data_by_sample = self.ignore_samples(data_by_sample)
        if not data_by_sample:
            return set()

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Save data file
        self.write_data_file(data_by_sample, "dragen_trimmer_metrics")

        table_data = DragenTrimmerMetrics.__get_table_data(data_by_sample)
        self.add_section(
            name="Trimmer Metrics",
            anchor="trimmer-metrics",
            description="""
            Metrics on trimmed reads.
            """,
            plot=table.plot(
                table_data,
                pconfig={
                    "id": "dragen-trimmer-metrics-table",
                    "namespace": "Trimmer Metrics",
                    "title": "DRAGEN: Trimmer Metrics",
                },
            ),
        )

        return data_by_sample.keys()

    @staticmethod
    def __get_table_data(data_by_sample: Dict) -> Dict:
        ANALYSIS = "TRIMMER STATISTICS"
        trimmer_data = {}
        for sample, data in data_by_sample.items():
            trimmer_data[sample] = {metric: stat for metric, stat in data[ANALYSIS].items()}
        return trimmer_data


def parse_trimmer_metrics_file(f):
    """
    sample.trimmer_metrics.csv

    TRIMMER STATISTICS,,Total input reads,18531840
    TRIMMER STATISTICS,,Total input bases,2779776000
    TRIMMER STATISTICS,,Total input bases R1,1389888000
    TRIMMER STATISTICS,,Total input bases R2,1389888000
    TRIMMER STATISTICS,,Average input read length,150
    TRIMMER STATISTICS,,Total trimmed reads,0,0.00
    TRIMMER STATISTICS,,Total trimmed bases,0,0.00
    """

    data = defaultdict(dict)
    for line in f["f"].splitlines():
        tokens = line.split(",")
        if len(tokens) == 4:
            analysis, _, metric, stat = tokens
            try:
                stat = int(stat)
            except ValueError:
                try:
                    stat = float(stat)
                except ValueError:
                    pass
            val = stat
        elif len(tokens) == 5:
            analysis, _, metric, stat, percentage = tokens
            try:
                percentage = float(percentage)
            except ValueError:
                pass
            val = percentage
        else:
            log.error(f"DRAGEN trimming metrics: unexpected number of tokens in line {line}")
            continue
        data[analysis][metric] = val

    return data



#!/usr/bin/env python

""" Super Special-Case MultiQC module to produce report section on MultiQC run time """


import logging

from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import bargraph
from multiqc.utils import report

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Run Time",
            anchor="multiqc_runtime",
            info="""
                This analysis is about the MultiQC run itself, profiling the time spent
                on different parts of the MultiQC execution. It is designed to help
                developers optimise how they run MultiQC, to get the most efficient
                and fastest configuration possible. For more information, see the
                <a href="https://multiqc.info/docs/#optimising-run-time" target="_blank">MultiQC documentation</a>.
            """,
        )

        log.info("Running run time profiling module")

        self.file_search_stats_section()

        self.search_pattern_times_section()

    def file_search_stats_section(self):
        """Count of all files iterated through by MultiQC, by category"""

        pdata = dict()
        pcats = dict()
        for key in sorted(report.file_search_stats, key=report.file_search_stats.get, reverse=True):
            if "skipped_" in key:
                s_name = f"Skipped: {key.replace('skipped_', '').replace('_', ' ').capitalize()}"
                pcats[key] = {"name": key, "color": "#999999"}
            else:
                s_name = key
                pcats[key] = {"name": key, "color": "#7cb5ec"}
            pdata[s_name] = {key: report.file_search_stats[key]}

        pconfig = {
            "id": "multiqc_runtime_files_searched_plot",
            "title": "MultiQC: Files searched",
            "ylab": "Number of files",
            "use_legend": False,
            "cpswitch": False,
        }

        self.add_section(
            name="Files searched",
            anchor="multiqc_runtime_files_searched",
            description="""
                Number of files searched by MultiQC, categorised by what happened to them.
                **Total file searches: {}**.
            """.format(sum(report.file_search_stats.values())),
            helptext="""
                Note that only files are considered in this plot - skipped directories are not shown.

                Some search patterns do not discard files after they match (`shared: true`), so it is possible
                that some files may be double-counted in this plot.

                * `Skipped: No match` - File was searched, but didn't match any search patterns
                * `Skipped: Ignore pattern` - File matched a MultiQC ignore pattern (see `-x` / `--ignore` / `config.fn_ignore_paths`)
                * `Skipped: Filesize limit` - File was skipped because it was too large (see `config.log_filesize_limit`)
                * `Skipped: Symlinks` - File was a symlink and skipped (see `config.ignore_symlinks`)
                * `Skipped: Not a file` - File could not be read (eg. was a unix pipe or something)
            """,
            plot=bargraph.plot(pdata, pcats, pconfig),
        )

    def search_pattern_times_section(self):
        """Section with a bar plot showing the time spent on each search pattern"""

        pdata = dict()
        for key in sorted(report.runtimes["sp"], key=report.runtimes["sp"].get, reverse=True):
            pdata[key] = {"time": report.runtimes["sp"][key]}

        pconfig = {
            "id": "multiqc_runtime_search_patterns_plot",
            "title": "MultiQC: Time per search pattern key",
            "ylab": "Time (seconds)",
            "use_legend": False,
            "cpswitch": False,
        }

        self.add_section(
            name="Search patterns",
            anchor="multiqc_runtime_search_patterns",
            description="""
                Time spent running each search pattern to find files for MultiQC modules.
                **Total file search time: {:.2f} seconds**.
            """.format(report.runtimes["total_sp"]),
            helptext="""
                **NOTE: Usually, MultiQC run time is fairly insignificant - in the order of seconds.
                Unless you are running MultiQC on many thousands of analysis files, optimising this process
                will have limited practical benefit.**

                MultiQC works by recursively looking through all files found in the analysis directories.
                After skipping any that are too big / binary file types etc, it uses the search patterns
                defined in `multiqc/utils/search_patterns.yaml`.
                These work by matching either file names or file contents. Generally speaking, matching
                filenames is super fast and matching file contents is slower.

                Please see the [MultiQC Documentation](https://multiqc.info/docs/#optimising-run-time)
                for information on how to optimise MultiQC to speed this process up.
                The plot below shows which search keys are running and how long each has taken to run in
                total. This should help to guide you to where optimisation is most worthwhile.
            """,
            plot=bargraph.plot(pdata, None, pconfig),
        )



""" MultiQC module to parse output from Bamtools """

import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

# Import the Bamtools submodules
from . import stats

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """Bamtools is a collection of scripts. This MultiQC module
    supports some but not all. The code for each script is split
    into its own file and adds a section to the module ooutput if
    logs are found."""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Bamtools",
            anchor="bamtools",
            href="https://github.com/pezmaster31/bamtools",
            info="provides both a programmer's API and an end-user's toolkit for handling BAM files.",
            doi="10.1093/bioinformatics/btr174",
        )

        # Set up class objects to hold parsed data
        self.general_stats_headers = dict()
        self.general_stats_data = dict()
        n = dict()

        # Call submodule functions
        n["stats"] = stats.parse_reports(self)
        if n["stats"] > 0:
            log.info(f"Found {n['stats']} bamtools stats reports")

        # Exit if we didn't find anything
        if sum(n.values()) == 0:
            raise ModuleNoSamplesFound

        # Add to the General Stats table (has to be called once per MultiQC module)
        self.general_stats_addcols(self.general_stats_data, self.general_stats_headers)



""" MultiQC module to parse TsTv by alternative allele count from vcftools TsTv-by-count """

import logging

from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


class TsTvByCountMixin:
    def parse_tstv_by_count(self):
        """Create the HTML for the TsTv by alternative allele count linegraph plot."""

        self.vcftools_tstv_by_count = dict()
        for f in self.find_log_files("vcftools/tstv_by_count", filehandles=True):
            d = {}
            for line in f["f"].readlines()[1:]:  # don't add the header line (first row)
                key = float(line.split()[0])  # taking the first column (alternative allele count) as key
                val = float(line.split()[3])  # taking Ts/Tv as value
                d[key] = val
            self.vcftools_tstv_by_count[f["s_name"]] = d
            self.add_data_source(f, "TsTv by counts")

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, f["s_name"])

        # Filter out ignored sample names
        self.vcftools_tstv_by_count = self.ignore_samples(self.vcftools_tstv_by_count)

        if len(self.vcftools_tstv_by_count) == 0:
            return 0

        # Write data to file
        self.write_data_file(self.vcftools_tstv_by_count, "vcftools_tstv_by_count")

        pconfig = {
            "id": "vcftools_tstv_by_count",
            "title": "VCFTools: TsTv by Count",
            "ylab": "TsTv Ratio",
            "xlab": "Alternative Allele Count",
            "xmin": 0,
            "ymin": 0,
            "smooth_points": 400,  # this limits huge filesizes and prevents browser crashing
            "smooth_points_sumcounts": False,
        }

        helptext = """
        `Transition` is a purine-to-purine or pyrimidine-to-pyrimidine point mutations.
        `Transversion` is a purine-to-pyrimidine or pyrimidine-to-purine point mutation.
        `Alternative allele count` is the number of alternative alleles at the site.
        Note: only bi-allelic SNPs are used (multi-allelic sites and INDELs are skipped.)
        Refer to Vcftools's manual (https://vcftools.github.io/man_latest.html) on `--TsTv-by-count`
        """

        self.add_section(
            name="TsTv by Count",
            anchor="vcftools-tstv-by-count",
            description="Plot of `TSTV-BY-COUNT` - the transition to transversion ratio as a function of alternative allele count from the output of vcftools TsTv-by-count.",
            helptext=helptext,
            plot=linegraph.plot(self.vcftools_tstv_by_count, pconfig),
        )

        return len(self.vcftools_tstv_by_count)



""" MultiQC module to parse output from odgi stats """


import logging

import yaml

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, table

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """graph
    The MultiQC module to parse and plot odgi stats output.
    """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="ODGI",
            anchor="odgi",
            href="https://github.com/pangenome/odgi",
            info="is an optimized dynamic graph/genome implementation, for efficient analysis and manipulation of pangenome graphs structured in the variation graph model.",
            # Can't find a DOI // doi=
        )

        # Parse odgi stats data
        self.odgi_stats_map = dict()
        for f in self.find_log_files("odgi", filehandles=True):
            self.parse_odgi_stats_report(f)

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, f["s_name"])

        self.odgi_stats_map = self.ignore_samples(self.odgi_stats_map)

        # No samples found
        if len(self.odgi_stats_map) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.odgi_stats_map)} report{'s' if len(self.odgi_stats_map) > 1 else ''}")

        # Write parsed report data to a file
        self.write_data_file(self.odgi_stats_map, "multiqc_odgi_stats")

        # Plot detailed odgi stats in an extra section
        self.odgi_stats_table()

        # Plot the odgi stats metrics as a lineplot
        self.plot_sum_of_path_nodes_distances()
        self.plot_mean_links_length()

    def parse_odgi_stats_report(self, f):
        """
        Load the odgi stats YAML file.
        The stats reports assume the following YAML structure:

        ---
        length: 206263
        nodes: 3751
        edges: 5195
        paths: 13
        num_weakly_connected_components: 1
        weakly_connected_components:
          - component:
              id: 0
              nodes: 3751
              is_acyclic: 'no'
        num_nodes_self_loops:
          total: 0
          unique: 0
        A: 57554
        C: 43275
        G: 41944
        T: 63490
        mean_links_length:
          - length:
              path: all_paths
              in_node_space: 1.64973
              in_nucleotide_space: 7.34035
              num_links_considered: 202793
              num_gap_links_not_penalized: 147940
        sum_of_path_node_distances:
          - distance:
              path: all_paths
              in_node_space: 5.53383
              in_nucleotide_space: 2.1454
              nodes: 202806
              nucleotides: 3757597
              num_penalties: 231
              num_penalties_different_orientation: 0
        """
        # Load the YAML file
        try:
            data = yaml.load(f["f"], Loader=yaml.SafeLoader)
        except Exception as e:
            log.warning(f"Could not parse YAML for '{f}': \n  {e}")
            return

        # Flatten the data structure
        data_flat = self.compress_stats_data(data)

        # Calculate additional metrics
        try:
            data_flat["pct_gc"] = ((data_flat["G"] + data_flat["C"]) / data_flat["length"]) * 100.0
            data_flat["pct_n"] = (data_flat["N"] / data_flat["length"]) * 100.0
        except ZeroDivisionError:
            data_flat["pct_gc"] = 0.0
            data_flat["pct_n"] = 0.0

        # Clean up the sample name
        s_name = self.extract_sample_name(f["fn"])
        s_name = self.clean_s_name(s_name, f)

        # Add to the data sources file
        self.add_data_source(f, s_name)

        # Add the data to the main odgi_stats_map dict
        if s_name in self.odgi_stats_map:
            log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
        self.odgi_stats_map[s_name] = data_flat

    def odgi_stats_table(self):
        """
        Detailed odgi stats in this extra table.
        """
        headers = {
            "length": {
                "title": "Length",
                "description": "Graph length in nucleotides.",
                "scale": "BuPu",
                "format": "{:,.0f}",
            },
            "nodes": {
                "title": "Nodes",
                "description": "Number of nodes in the graph.",
                "scale": "OrRd",
                "format": "{:,.0f}",
            },
            "edges": {
                "title": "Edges",
                "description": "Number of edges in the graph.",
                "scale": "PuBu",
                "format": "{:,.0f}",
            },
            "paths": {
                "title": "Paths",
                "description": "Number of paths in the graph.",
                "scale": "Greens",
                "format": "{:,.0f}",
            },
            "components": {
                "title": "Components",
                "description": "Number of weakly connected components in the graph.",
                "scale": "Oranges",
                "format": "{:,.0f}",
            },
            "A": {
                "title": "A",
                "description": "Number of adenine bases in the graph.",
                "scale": "Spectral",
                "format": "{:,.0f}",
                "shared_key": "nucleotides",
            },
            "C": {
                "title": "C",
                "description": "Number of cytosine bases in the graph.",
                "scale": "Greys",
                "format": "{:,.0f}",
                "shared_key": "nucleotides",
            },
            "T": {
                "title": "T",
                "description": "Number of thymine bases in the graph.",
                "scale": "Blues",
                "format": "{:,.0f}",
                "shared_key": "nucleotides",
            },
            "G": {
                "title": "G",
                "description": "Number of guanine bases in the graph.",
                "scale": "RdPu",
                "format": "{:,.0f}",
                "shared_key": "nucleotides",
            },
            "N": {
                "title": "N",
                "description": "Number of `N` basis in the graph.",
                "scale": "BuGn",
                "format": "{:,.0f}",
                "shared_key": "nucleotides",
            },
            "total": {
                "title": "Self Loops Nodes",
                "description": "Total number of nodes having self loops in the graph.",
                "scale": "YlGn",
                "hidden": True,
                "format": "{:,.0f}",
            },
            "unique": {
                "title": "Unique Self Loops Nodes",
                "description": "Number of unique nodes having self loops in the graph.",
                "scale": "BuGn",
                "hidden": True,
                "format": "{:,.0f}",
            },
            "pct_gc": {
                "title": "% GC",
                "description": "Percent of G/C bases in the graph.",
                "scale": "Spectral",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "hidden": True,
            },
            "pct_n": {
                "title": "% N",
                "description": "Percent of N bases in the graph.",
                "scale": "Reds",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "hidden": True,
            },
        }
        # Some of the headers are quite general and can clash with other modules.
        # Prepend odgi_ to keep them unique
        prefix_headers = dict()
        prefix_data = {}
        for h, v in headers.items():
            prefix_headers[f"odgi_{h}"] = v
        for s_name, d in self.odgi_stats_map.items():
            prefix_data[s_name] = {}
            for h in headers:
                prefix_data[s_name][f"odgi_{h}"] = d[h]
        tconfig = {
            "id": "odgi_table",
            "namespace": "ODGI",
            "table_title": "ODGI Stats",
        }
        self.add_section(
            name="Detailed ODGI stats table.",
            anchor="extended_odgi_stats",
            plot=table.plot(prefix_data, prefix_headers, tconfig),
        )

    def plot_sum_of_path_nodes_distances(self):
        """
        Plot odgi path node distances bar plot.
        """

        cats = [
            {
                "sum_of_path_nodes_distances_in_node_space": {
                    "name": "Sum of distances: node space",
                }
            },
            {
                "sum_of_path_nodes_distances_in_nucleotide_space": {
                    "name": "Sum of distances: nucleotide space",
                }
            },
        ]
        pconfig = {
            "id": "odgi_sum_of_path_nodes_distances_plot",
            "title": "ODGI: Sum of path node distances",
            "ylab": "Distance",
            "yDecimals": False,
            "cpswitch": False,
            "logswitch": True,
            "logswitch_active": True,
            "data_labels": [
                {"name": "Node space", "ylab": "Distance"},
                {"name": "Nucleotide space", "ylab": "Distance"},
            ],
        }
        self.add_section(
            name="Sum of path node distances",
            anchor="odgi_sum_of_path_nodes_distances",
            description="""
                For each path we iterate from node to node and count the node / nucleotide distance of nodes on the pangenome
                level normalized by the path length. If a node is reversed, we count the node distance twice.
            """,
            helptext="""
                This value allows you to evaluate the sorting goodness - how linear the graph is.
            """,
            plot=bargraph.plot([self.odgi_stats_map, self.odgi_stats_map], cats, pconfig),
        )

    def plot_mean_links_length(self):
        """
        Plot odgi path node distances bar plot.
        """
        cats = [
            {
                "mean_links_length_in_node_space": {
                    "name": "Mean links length: node space",
                }
            },
            {
                "mean_links_length_in_nucleotide_space": {
                    "name": "Mean links length: nucleotide space",
                }
            },
        ]
        pconfig = {
            "id": "odgi_mean_links_length_plot",
            "title": "ODGI: Mean links length",
            "ylab": "Distance",
            "yDecimals": False,
            "cpswitch": False,
            "logswitch": True,
            "logswitch_active": True,
            "data_labels": [
                {"name": "Node space", "ylab": "Distance"},
                {"name": "Nucleotide space", "ylab": "Distance"},
            ],
        }

        self.add_section(
            name="Mean links length",
            anchor="odgi_mean_links_length",
            description="""
                For each path we iterate from node to node and count the node / nucleotide distance `mean_links_length`
                of nodes _within the same path only._ We then normalized by the path length.
            """,
            helptext="""
                This value allows you to evaluate the sorting goodness - how linear the graph is.
            """,
            plot=bargraph.plot([self.odgi_stats_map, self.odgi_stats_map], cats, pconfig),
        )

    @staticmethod
    def extract_sample_name(file_name):
        """
        Extracts the sample name from a given file name.
        Expects and returns one of seqwish, smooth or cons@*.
        If none of the above keywords are present, the full filename except for the ending is returned.
        """
        if "cons@" in file_name:
            file_name = file_name.split(".")
            consensus_identifier = list((e for e in file_name if "cons@" in e))
            return consensus_identifier[0]
        elif "smooth" in file_name:
            return "smooth"
        elif "seqwish" in file_name:
            return "seqwish"
        else:
            return ".".join(file_name.split(".")[:-3])

    @staticmethod
    def compress_stats_data(data) -> dict:
        """
        Compress odgi stats into a single dictionary to visualize.
        """
        mean_links_length = data["mean_links_length"]
        # we have to find the entry with path: 'all_paths', because odgi stats could emit a list of path names
        for line in mean_links_length:
            if line["length"]["path"] == "all_paths":
                length = line["length"]
        sum_of_path_nodes_distances = data["sum_of_path_node_distances"]
        # we have to find the entry with path: 'all_paths', because odgi stats could emit a list of path names
        for d in sum_of_path_nodes_distances:
            if d["distance"]["path"] == "all_paths":
                distance = d["distance"]
        return {
            "length": float(data["length"]),
            "nodes": float(data["nodes"]),
            "edges": float(data["edges"]),
            "paths": float(data["paths"]),
            "components": float(data["num_weakly_connected_components"]),
            "A": float(data["A"]),
            "C": float(data["C"]),
            "T": float(data["T"]),
            "G": float(data["G"]),
            "N": float(data.get("N", 0)),
            "total": float(data["num_nodes_self_loops"]["total"]),
            "unique": float(data["num_nodes_self_loops"]["unique"]),
            "mean_links_length_path": length["path"],
            "mean_links_length_in_node_space": length["in_node_space"],
            "mean_links_length_in_nucleotide_space": length["in_nucleotide_space"],
            "mean_links_length_num_links_considered": length["num_links_considered"],
            "mean_links_length_num_gap_links_not_penalized": length["num_gap_links_not_penalized"],
            "sum_of_path_nodes_distances_path": distance["path"],
            "sum_of_path_nodes_distances_in_node_space": distance["in_node_space"],
            "sum_of_path_nodes_distances_in_nucleotide_space": distance["in_nucleotide_space"],
            "sum_of_path_nodes_distances_nodes": distance["nodes"],
            "sum_of_path_nodes_distances_nucleotides": distance["nucleotides"],
            "sum_of_path_nodes_distances_num_penalties": distance["num_penalties"],
            "sum_of_path_nodes_distances_num_penalties_different_orientation": distance[
                "num_penalties_different_orientation"
            ],
        }



""" MultiQC module to parse output from Samblaster """


import logging
import os
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)

VERSION_REGEX = r"Version\ (\d{1}.\d+.\d+)"


class MultiqcModule(BaseMultiqcModule):
    """Samblaster"""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Samblaster",
            anchor="samblaster",
            href="https://github.com/GregoryFaust/samblaster",
            info="is a tool to mark duplicates and extract discordant and split reads from sam files.",
            doi="10.1093/bioinformatics/btu314",
        )

        self.samblaster_data = dict()
        for f in self.find_log_files("samblaster", filehandles=True):
            self.parse_samblaster(f)

        # Filter to strip out ignored sample names
        self.samblaster_data = self.ignore_samples(self.samblaster_data)

        if len(self.samblaster_data) == 0:
            raise ModuleNoSamplesFound

        headers = {
            "pct_dups": {
                "title": "% Dups",
                "description": "Percent Duplication",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "OrRd",
            }
        }

        self.general_stats_addcols(self.samblaster_data, headers)

        # Write parsed report data to a file
        self.write_data_file(self.samblaster_data, "multiqc_samblaster")

        log.info(f"Found {len(self.samblaster_data)} reports")

        self.add_barplot()

    def add_barplot(self):
        """Generate the Samblaster bar plot."""
        cats = {
            "n_nondups": {"name": "Non-duplicates"},
            "n_dups": {"name": "Duplicates"},
        }

        pconfig = {
            "id": "samblaster_duplicates",
            "title": "Samblaster: Number of duplicate reads",
            "ylab": "Number of reads",
        }
        self.add_section(plot=bargraph.plot(self.samblaster_data, cats, pconfig))

    def parse_samblaster(self, f):
        """Go through log file looking for samblaster output.
        Grab the name from the RG tag of the preceding bwa command"""

        # Should capture the following:
        # samblaster: Marked     1134898 of   43791982 (2.592%) total read ids as duplicates using 753336k \
        # memory in 1M1S(60.884S) CPU seconds and 3M53S(233S) wall time.
        dups_regex = (
            r"samblaster: (Removed|Marked)\s+(\d+)\s+of\s+(\d+) \((\d+.\d+)%\)\s*(total)?\s*read ids as duplicates"
        )

        input_file_regex = r"samblaster: Opening (\S+) for read."
        rgtag_name_regex = r"\\\\tID:(\S*?)\\\\t"
        data = {}
        s_name = None
        version = None
        fh = f["f"]
        for line in fh:
            match = re.search(VERSION_REGEX, line)
            if match is not None:
                version = match.group(1)
            # try to find name from RG-tag. If bwa mem is used upstream samblaster with pipes, then the bwa mem command
            # including the read group will be written in the log
            match = re.search(rgtag_name_regex, line)
            if match:
                s_name = self.clean_s_name(match.group(1), f)

            # try to find name from the input file name, if used
            match = re.search(input_file_regex, line)
            if match:
                basefn = os.path.basename(match.group(1))
                fname, ext = os.path.splitext(basefn)
                # if it's stdin, then try bwa RG-tag instead
                if fname != "stdin":
                    s_name = self.clean_s_name(fname, f)

            match = re.search(dups_regex, line)
            if match:
                data["n_dups"] = int(match.group(2))
                data["n_tot"] = int(match.group(3))
                data["n_nondups"] = data["n_tot"] - data["n_dups"]
                data["pct_dups"] = float(match.group(4))

        if s_name is None:
            s_name = f["s_name"]

        if version is not None:
            self.add_software_version(version, s_name)

        if len(data) > 0:
            if s_name in self.samblaster_data:
                log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")
            self.add_data_source(f, s_name)
            self.samblaster_data[s_name] = data



""" MultiQC module to parse output from Bamdst """


import logging
from collections import defaultdict
from typing import Dict, Union
import csv
import math
import os
import fnmatch

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import table, bargraph, linegraph
from multiqc.utils import config

# Initialise the logger
log = logging.getLogger(__name__)


default_exclude_contigs = [
    "*_alt",
    "*_decoy",
    "*_random",
    "*_fix",
    "HLA*",
    "chrUn*",
    "chrEBV",
    "chrM",
]


def _read_config():
    cfg = getattr(config, "bamdst", dict())
    if not isinstance(cfg, dict):
        return {}

    cfg["exclude_include_are_set_by_user"] = "exclude_contigs" in cfg or "include_contigs" in cfg

    cfg["include_contigs"] = cfg.get("include_contigs", [])
    if not isinstance(cfg["include_contigs"], list):
        cfg["include_contigs"] = []

    cfg["exclude_contigs"] = cfg.get("exclude_contigs", default_exclude_contigs)
    if not isinstance(cfg["exclude_contigs"], list):
        cfg["exclude_contigs"] = []

    if cfg["include_contigs"]:
        log.debug(f"Configuration: would try to include only these contigs: {', '.join(cfg['include_contigs'])}")
    if cfg["exclude_contigs"]:
        log.debug(
            f"Configuration: would exclude these contigs, unless all would be filtered: {', '.join(cfg['exclude_contigs'])}"
        )

    cutoff = cfg.get("perchrom_fraction_cutoff", 0.0)
    try:
        cutoff = float(cutoff)
    except ValueError:
        cutoff = 0.0
    if cutoff != 0.0:
        log.debug(
            f"Configuration: coverage cutoff of contigs is {cutoff * 100.0}%, unless everything would be filtered"
        )
    cfg["perchrom_fraction_cutoff"] = cutoff

    return cfg


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        super(MultiqcModule, self).__init__(
            name="Bamdst",
            anchor="bamdst",
            href="https://https://github.com/shiquan/bamdst",
            info="""is a lightweight tool to stat the depth coverage of target regions of bam file(s)""",
            # doi="", # No DOI
        )

        self.cfg = _read_config()
        data_by_sample = dict()
        data_by_chromosome_by_sample = dict()
        for f in self.find_log_files("bamdst/coverage"):
            if data := self._parse_coverage_report(f):
                data_by_sample[f["s_name"]] = data

                # Parse "chromosomes.report" sitting next to the "coverage.report" file:
                if os.path.exists(chroms_path := os.path.join(f["root"], "chromosomes.report")):
                    if data := self._parse_chromosomes_report(chroms_path, f["s_name"]):
                        data_by_chromosome_by_sample[f["s_name"]] = data

        data_by_sample = self.ignore_samples(data_by_sample)
        data_by_chromosome_by_sample = self.ignore_samples(data_by_chromosome_by_sample)
        if len(data_by_sample) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(data_by_sample)} reports")
        self.write_data_file(data_by_sample, "multiqc_bamdst_coverage")
        self.write_data_file(data_by_chromosome_by_sample, "multiqc_bamdst_chromosomes")

        self.build_tables(data_by_sample)
        self._build_per_chrom_plot(data_by_chromosome_by_sample)

    def _parse_coverage_report(self, f: Dict) -> Dict:
        """
        Parse one coverage report.
        """
        data = dict()
        s_name = None
        version = None
        for line in f["f"].split("\n"):
            line = line.strip()
            if line.startswith("## Files : "):
                # Split a line like "## Files : example/test1.bam example/test2.bam "
                # Cannot split by spaces because the file names can contain spaces as well
                path_line = line.split(":")[1].strip()
                for ext in [".bam", ".cram", ".sam"]:
                    path_line = path_line.replace(ext, "<EXT>")
                names = [
                    self.clean_s_name(sn.strip())
                    for sn in path_line.split("<EXT>")
                    if sn
                    if sn.strip() not in ["-", "/dev/stdin"]
                ]
                if names:
                    f["s_name"] = "-".join(names)  # concatenating file names
                continue
            if line.startswith("## Version : "):
                version = line.split(":")[1].strip()
                continue
            fields = line.split("\t")
            if not len(fields) == 2:
                continue
            metric, value = fields
            if value.endswith("%"):
                value = value.strip("%")
                try:
                    value = float(value)
                except ValueError:
                    continue
                value = value / 100.0
            else:
                try:
                    value = int(value)
                except ValueError:
                    try:
                        value = float(value)
                    except ValueError:
                        continue
            data[metric] = value

        if s_name and version and data:
            self.add_software_version(version, sample=s_name)
            self.add_data_source(f, s_name=s_name, section="coverage table")
        return data

    def _parse_chromosomes_report(self, path: str, s_name: str) -> Dict[str, Dict[str, Union[float, str]]]:
        data_by_contig = defaultdict(dict)
        with open(path) as fh:
            reader: csv.DictReader = csv.DictReader(fh, delimiter="\t")
            for row in reader:
                contig = row["#Chromosome"].strip()
                for k, v in row.items():
                    if k == "#Chromosome":
                        continue
                    v = v.strip()
                    try:
                        v = float(v)
                    except ValueError:
                        pass
                    data_by_contig[contig][k.strip()] = v

        if data_by_contig:
            self.add_data_source(source=path, s_name=s_name, section="per-chromosome")
        return data_by_contig

    def build_tables(self, data_by_sample):
        headers = {
            "[Total] Mapped Reads": {
                "title": f"Mapped ({config.read_count_prefix})",
                "description": f"Mapped reads ({config.read_count_desc})",
                "scale": "RdYlGn",
                "min": 0,
                "modify": lambda x: x * config.read_count_multiplier,
                "shared_key": "read_count",
            },
            "[Total] Fraction of Mapped Reads": {
                "title": "Mapped",
                "description": f"Fraction of mapped reads in all reads ({config.read_count_desc})",
                "scale": "PuBu",
                "min": 0.0,
                "max": 100.0,
                "suffix": "%",
                "modify": lambda x: x * 100.0,
            },
            "[Target] Target Reads": {
                "title": f"Target reads ({config.read_count_prefix})",
                "description": f"Reads mapped on target ({config.read_count_desc})",
                "scale": "RdYlGn",
                "min": 0,
                "modify": lambda x: x * config.read_count_multiplier,
            },
            "[Target] Fraction of Target Reads in all reads": {
                "title": "% in all",
                "description": "Fraction of target reads in all reads",
                "scale": "PuBu",
                "min": 0.0,
                "max": 100.0,
                "suffix": "%",
                "modify": lambda x: x * 100.0,
            },
            "[Target] Fraction of Target Reads in mapped reads": {
                "title": "% in mapped",
                "description": "Fraction of target reads in mapped reads",
                "scale": "PuBu",
                "min": 0.0,
                "max": 100.0,
                "suffix": "%",
                "modify": lambda x: x * 100.0,
            },
            "[Target] Average depth": {
                "title": "Avg depth",
                "description": "Average depth, on target",
                "min": 0,
                "suffix": "X",
                "scale": "BuPu",
            },
            "[Target] Len of region": {
                "title": "Region len",
                "description": "Length of target region",
                "min": 0,
                "format": "{:,d}",
                "suffix": " bp",
                "scale": "Greys",
            },
        }

        coverage_metrics = []
        for s_name, d in data_by_sample.items():
            for k in d.keys():
                if k.startswith("[Target] Coverage (") and k not in coverage_metrics:
                    coverage_metrics.append(k)
        for m in coverage_metrics:
            short = m.replace("[Target] Coverage (", "").replace(")", "")
            headers[m] = {
                "title": short,
                "description": f"Target coverage ({short})",
                "min": 0,
                "suffix": "%",
                "format": "{:.2f}",
                "scale": "RdYlGn",
                "modify": lambda x: x * 100.0,
            }

        self.add_section(
            name="Coverage metrics",
            anchor="bamdst-coverage",
            plot=table.plot(
                data_by_sample,
                headers,
                pconfig={
                    "id": "bamdst-coverage-table",
                    "title": "Bamdst: coverage metrics",
                },
            ),
        )

        genstats_headers = {
            k: v
            for k, v in headers.items()
            if k
            in [
                "[Target] Average depth",
                "[Target] Coverage (>0x)",
            ]
        }
        self.general_stats_addcols(data_by_sample, genstats_headers)

    def _filter_contigs(self, data_by_chrom_by_sample):
        # Filter contigs based on include/exclude patterns
        filtered_data_by_chrom_by_sample = defaultdict(dict)
        for s_name, data_by_chrom in data_by_chrom_by_sample.items():
            for contig, data in data_by_chrom.items():
                if any(fnmatch.fnmatch(contig, str(pattern)) for pattern in self.cfg["exclude_contigs"]):
                    try:
                        if self.cfg.get("show_excluded_debug_logs") is True:
                            log.debug(f"Skipping excluded contig '{contig}'")
                    except (AttributeError, KeyError):
                        pass
                    continue

                # filter out contigs based on inclusion patterns
                if len(self.cfg["include_contigs"]) > 0 and not any(
                    fnmatch.fnmatch(contig, pattern) for pattern in self.cfg["include_contigs"]
                ):
                    # Commented out since this could be many thousands of contigs!
                    # log.debug(f"Skipping not included contig '{contig}'")
                    continue
                filtered_data_by_chrom_by_sample[s_name][contig] = data
        if not filtered_data_by_chrom_by_sample:
            filtered_data_by_chrom_by_sample = data_by_chrom_by_sample
            (log.warning if self.cfg.get("exclude_include_are_set_by_user", False) else log.debug)(
                "All contigs would be filtered out by the contigs inclusion/exclusion patterns. "
                "Consider changing the contig.include_contigs/contig.exclude_contigs parameters. "
                "Keeping all contigs."
            )
        data_by_chrom_by_sample = filtered_data_by_chrom_by_sample

        # Filter contigs based on coverage cutoff. Contigs only will be excluded if they
        # fail the cutoff in _all_ samples.
        filtered_data_by_chrom_by_sample = defaultdict(dict)
        if (cutoff := self.cfg.get("perchrom_fraction_cutoff", 0)) > 0:
            passing_contigs = set()
            for s_name, data_by_chrom in data_by_chrom_by_sample.items():
                for contig, data in data_by_chrom.items():
                    pct = data.get("DATA(%)")
                    if pct is not None and isinstance(pct, float) and pct > cutoff:
                        passing_contigs.add(contig)

            rejected_contigs = set()
            for s_name, perchrom in data_by_chrom_by_sample.items():
                for contig, data in perchrom.items():
                    if contig not in passing_contigs:
                        rejected_contigs.add(contig)
                    else:
                        filtered_data_by_chrom_by_sample[s_name][contig] = data

            if rejected_contigs:
                if self.cfg.get("show_excluded_debug_logs") is True:
                    log.debug(
                        f"Skipping {len(rejected_contigs)} contigs not passing the "
                        f"cutoff of {self.cfg['perchrom_fraction_cutoff']}%:. "
                        f"Skipped contigs: {''.join(rejected_contigs)}"
                    )
            if not filtered_data_by_chrom_by_sample:
                filtered_data_by_chrom_by_sample = data_by_chrom_by_sample
                log.warning(
                    f"All contigs would be filtered out by the cutoff of "
                    f"{self.cfg['perchrom_fraction_cutoff']}%. Consider changing the cut-off. "
                    f"Keeping all contigs."
                )
            data_by_chrom_by_sample = filtered_data_by_chrom_by_sample

        return data_by_chrom_by_sample

    def _build_per_chrom_plot(self, data_by_chrom_by_sample):
        def _chrom_key(name):
            k = "".join(c for c in name if c.isdigit())
            return int(k) if k else math.inf

        # Sort chromosomes by name. Sort smartly: chr2 should go before chr10
        for s_name, data_by_chrom in data_by_chrom_by_sample.items():
            data_by_chrom_by_sample[s_name] = dict(sorted(data_by_chrom.items(), key=lambda x: _chrom_key(x[0])))

        # Filter contigs
        filt_data_by_chrom_by_sample = self._filter_contigs(data_by_chrom_by_sample)
        contigs = set.union(*[set(d.keys()) for d in data_by_chrom_by_sample.values()])
        filt_contigs = set.union(*[set(d.keys()) for d in filt_data_by_chrom_by_sample.values()])
        if contigs == filt_contigs or len(filt_contigs) == 0:
            datasets = [data_by_chrom_by_sample]
            data_labels = None
        else:
            datasets = [filt_data_by_chrom_by_sample, data_by_chrom_by_sample]
            data_labels = ["Main contigs", "All contigs"]

        depth_datasets = []
        cov_datasets = []
        for dataset in datasets:
            # Extracting two metrics to plot: average depth and coverage percentage
            depth_datasets.append(defaultdict(dict))
            cov_datasets.append(defaultdict(dict))
            for s_name, data_by_chrom in dataset.items():
                for chrom, d in data_by_chrom.items():
                    depth_datasets[-1][s_name][chrom] = d["Avg depth"]
                    cov_datasets[-1][s_name][chrom] = d["Coverage%"]

        num_chroms = max([len(by_chrom.keys()) for by_chrom in depth_datasets[0].values()])
        if num_chroms > 1:
            perchrom_depth_plot = linegraph.plot(
                depth_datasets,
                pconfig={
                    "id": "bamdst-depth-per-contig-plot",
                    "title": "Bamdst: average depth per contig",
                    "xlab": "Region",
                    "ylab": "Average depth",
                    "categories": True,
                    "tt_decimals": 1,
                    "tt_suffix": "x",
                    "smooth_points": 500,
                    "logswitch": True,
                    "hide_zero_cats": False,
                    "ymin": 0,
                    "data_labels": data_labels,
                },
            )
            perchrom_cov_plot = linegraph.plot(
                cov_datasets,
                pconfig={
                    "id": "bamdst-cov-per-contig-plot",
                    "title": "Bamdst: coverage percentage of each contig",
                    "xlab": "Region",
                    "ylab": "Coverage %",
                    "categories": True,
                    "tt_decimals": 1,
                    "tt_suffix": "%",
                    "smooth_points": 500,
                    "logswitch": True,
                    "hide_zero_cats": False,
                    "ymax": 100,
                    "ymin": 0,
                    "data_labels": data_labels,
                },
            )
        else:
            perchrom_depth_plot = bargraph.plot(
                depth_datasets,
                pconfig={
                    "id": "bamdst-depth-per-contig-plot",
                    "title": "Bamdst: average depth",
                    "xlab": "Sample",
                    "ylab": "Average depth",
                    "tt_suffix": "x",
                    "hide_zero_cats": False,
                    "ymin": 0,
                    "data_labels": data_labels,
                },
            )
            perchrom_cov_plot = bargraph.plot(
                cov_datasets,
                pconfig={
                    "id": "bamdst-cov-per-contig-plot",
                    "title": "Bamdst: coverage percentage",
                    "xlab": "Sample",
                    "ylab": "Coverage %",
                    "tt_suffix": "%",
                    "hide_zero_cats": False,
                    "ymax": 100,
                    "ymin": 0,
                    "data_labels": data_labels,
                },
            )

        self.add_section(
            name="Average depth per chromosome",
            anchor="bamdst-depth-per-contig",
            plot=perchrom_depth_plot,
        )
        self.add_section(
            name="Coverage % of each contig",
            anchor="bamdst-coverage-per-contig",
            plot=perchrom_cov_plot,
        )
